#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from functools import partialmethod
from hashlib import sha1

import re
import json

from ..model.constants import CORE_CONTEXT
from .exceptions import NgsiApiError

if TYPE_CHECKING:
    from .client import Client
from .exceptions import NgsiResourceNotFoundError, rfc7807_error_handle


class Subscriptions:
    def __init__(self, client: Client, url: str):
        self._client = client
        self._session = client.session
        self.url = url

    @rfc7807_error_handle
    def create(self, subscr: dict, raise_on_conflict: bool = True) -> bool:
        if raise_on_conflict:
            conflicts = self.conflicts(subscr)
            if conflicts:
                raise ValueError(f"Some subscriptions already exist with same target : {conflicts}")
        r = self._session.post(f"{self.url}/", json=subscr)
        self._client.raise_for_status(r)
        location = r.headers.get("Location")
        if location is None:
            if self._client.ignore_errors:
                return None
            else:
                raise NgsiApiError("Missing Location header")
        id_returned_from_broker = location.rsplit("/", 1)[-1]
        id = subscr.get("id")
        if id is not None and id != id_returned_from_broker:
            raise NgsiApiError(
                f"Broker returned wrong id. Expected={id} Returned={id_returned_from_broker}"
            )
        return id_returned_from_broker

    @rfc7807_error_handle
    def list(self, pattern: str = None, ctx: str = CORE_CONTEXT) -> Optional[dict]:
        headers = {
            "Accept": "application/ld+json",
            "Content-Type": None,
        }  # overrides session headers
        if ctx is not None:
            headers[
                "Link"
            ] = f'<{ctx}>; rel="{CORE_CONTEXT}"; type="application/ld+json"'
        r = self._session.get(f"{self.url}")
        subscriptions = r.json()
        if pattern is not None:
            subscriptions = [
                x
                for x in subscriptions
                if re.search(
                    pattern, x.get("name", "") + x.get("description", ""), re.IGNORECASE
                )
            ]
        return subscriptions

    @staticmethod
    def _criteria_only(subscr: dict):
        params = subscr.copy()
        params.pop("id", None)
        params.pop("name", None)
        params.pop("description", None)
        params.pop("isActive", None)
        return params

    @staticmethod
    def _hash(subscr: dict):
        criteria = Subscriptions._criteria_only(subscr)
        return sha1(json.dumps(criteria, sort_keys=True).encode("utf-8")).digest()

    @rfc7807_error_handle
    def conflicts(self, subscr: dict, ctx: str = CORE_CONTEXT) -> list:
        hashref = Subscriptions._hash(subscr)
        headers = {
            "Accept": "application/ld+json",
            "Content-Type": None,
        }  # overrides session headers
        if ctx is not None:
            headers[
                "Link"
            ] = f'<{ctx}>; rel="{CORE_CONTEXT}"; type="application/ld+json"'
        r = self._session.get(f"{self.url}")
        return [x for x in r.json() if Subscriptions._hash(x) == hashref]

    @rfc7807_error_handle
    def get(self, id: str, ctx: str = CORE_CONTEXT) -> dict:
        headers = {
            "Accept": "application/ld+json",
            "Content-Type": None,
        }  # overrides session headers
        if ctx is not None:
            headers[
                "Link"
            ] = f'<{ctx}>; rel="{CORE_CONTEXT}"; type="application/ld+json"'
        r = self._session.get(f"{self.url}/{id}", headers=headers)
        self._client.raise_for_status(r)
        return r.json()

    @rfc7807_error_handle
    def exists(self, id: str, ctx: str = CORE_CONTEXT) -> bool:
        try:
            payload = self.get(id, ctx)
            if payload:
                return "@context" in payload
        except NgsiResourceNotFoundError:
            return False
        return False

    @rfc7807_error_handle
    def _delete(self, id: str) -> bool:
        r = self._session.delete(f"{self.url}/{id}")
        self._client.raise_for_status(r)
        return bool(r)

    @rfc7807_error_handle
    def delete(self, pattern: str) -> bool:
        deleted = False
        for subscription in self.list(pattern):
            deleted |= self._delete(subscription["id"])
        return deleted

    purge = partialmethod(delete, pattern=None)



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

from __future__ import annotations
from dataclasses import dataclass


@dataclass
class PostalAddress:
    """A PostalAddress as defined here : https://schema.org/PostalAddress.
    """
    country: str = None
    locality: str = None
    region: str = None
    pobox: str = None
    postalcode: str = None
    streetaddress: str = None

    def to_dict(self):
        d = {}
        if self.streetaddress and self.pobox:
            raise ValueError("Provide either a Street Address or a PO Box but not both")
        if self.streetaddress:
            d["streetAddress"] = self.streetaddress
        if self.pobox:
            d["postOfficeBoxNumber"] = self.pobox
        if self.locality:
            d["addressLocality"] = self.locality
        if self.postalcode:
            d["postalCode"] = self.postalcode
        if self.region:
            d["addressRegion"] = self.region
        if self.country:
            d["addressCountry"] = self.country
        if not d:
            raise ValueError("PostalAddress is empty")
        return d


class PostalAddressBuilder:
    """A helper class that allows to easily build a PostalAddress property.

    Example:
    --------
    >>> from ngsildclient import *
    >>> builder = PostalAddressBuilder()
    >>> address = builder.street("C/ La Pereda 14")
        .locality("Santander")
        .region("Cantabria")
        .country("Spain")
        .build()
    >>> # Add an address property to the entity you're creating
    >>> busstop = Entity("PublicTransportStop", "santander:busStop:463")
    >>> busstop.prop("adress", address)
    >>> busstop.pprint()
    {
        "@context": [
            "https://uri.etsi.org/ngsi-ld/v1/ngsi-ld-core-context.jsonld"
        ],
        "id": "urn:ngsi-ld:PublicTransportStop:santander:busStop:463",
        "type": "PublicTransportStop",
        "adress": {
            "type": "Property",
            "value": {
                "streetAddress": "C/ La Pereda 14",
                "addressLocality": "Santander",
                "addressRegion": "Cantabria",
                "addressCountry": "Spain"
            }
        }
    }
    """
    def __init__(self):
        self._addr: PostalAddress = PostalAddress()

    def country(self, value: str) -> PostalAddressBuilder:
        self._addr.country = value
        return self

    def locality(self, value: str) -> PostalAddressBuilder:
        self._addr.locality = value
        return self

    def region(self, value: str) -> PostalAddressBuilder:
        self._addr.region = value
        return self

    def pobox(self, value: str) -> PostalAddressBuilder:
        self._addr.pobox = value
        return self

    def postalcode(self, value: str) -> PostalAddressBuilder:
        self._addr.postalcode = value
        return self

    def street(self, value: str) -> PostalAddressBuilder:
        self._addr.streetaddress = value
        return self

    def build(self) -> dict:
        return self._addr.to_dict()



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

from __future__ import annotations
from typing import TYPE_CHECKING

import logging

if TYPE_CHECKING:
    from .client import Client

from .constants import *
from .exceptions import rfc7807_error_handle
from ..model.entity import Entity


logger = logging.getLogger(__name__)


class BatchOp:
    def __init__(self, client: Client, url: str):
        self._client = client
        self._session = client.session
        self.url = url

    @rfc7807_error_handle
    def create(
        self, entities: List[Entity], skip: bool = False, overwrite: bool = False
    ) -> tuple[bool, dict]:
        r = self._session.post(
            f"{self.url}/create/", json=[entity._payload for entity in entities]
        )
        if r.status_code == 201:
            return True, r.json()
        else:
            return False, r.json()

    @rfc7807_error_handle
    def upsert(self, entities: List[Entity]) -> tuple[bool, dict]:
        r = self._session.post(
            f"{self.url}/upsert/", json=[entity._payload for entity in entities]
        )
        if r.status_code == 201:
            return True, r.json()
        elif r.status_code == 204:
            return True, {
                "success": "all entities already existed and are successfully updated"
            }
        else:
            return False, r.json()

    @rfc7807_error_handle
    def update(self, entities: List[Entity]) -> tuple[bool, dict]:
        r = self._session.post(
            f"{self.url}/update/", json=[entity._payload for entity in entities]
        )
        if r.status_code == 204:
            return True, {"success": "all entities have been successfully updated"}
        else:
            return False, r.json()

    @rfc7807_error_handle
    def delete(self, entities: List[Entity]) -> tuple[bool, dict]:
        r = self._session.post(
            f"{self.url}/delete/", json=[entity.id for entity in entities]
        )
        if r.status_code == 204:
            return True, {
                "success": "all entities existed and have been successfully deleted"
            }
        else:
            return False, r.json()



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

from __future__ import annotations
from typing import TYPE_CHECKING, Optional

import logging

if TYPE_CHECKING:
    from .client import Client

from .constants import *
from .exceptions import (
    NgsiAlreadyExistsError,
    NgsiApiError,
    rfc7807_error_handle
)
from ..model.entity import Entity


logger = logging.getLogger(__name__)


class Entities:
    def __init__(self, client: Client, url: str):
        self._client = client
        self._session = client.session
        self.url = url

    @rfc7807_error_handle
    def create(
        self,
        entity: Entity,
        skip: bool = False,
        overwrite: bool = False
    ) -> Optional[Entity]:
        r = self._session.post(
            f"{self.url}/",
            json=entity._payload,
        )

        if r.status_code == 409:  # already exists
            if skip:
                return None
            elif overwrite or self._client.overwrite:
                return self.update(entity, check_exists=False)

        self._client.raise_for_status(r)
        location = r.headers.get("Location")
        if location is None:
            if self._client.ignore_errors:
                return None
            else:
                raise NgsiApiError("Missing Location header")
        logger.info(f"{r.status_code=}")
        logger.info(f"{location=}")
        id_returned_from_broker = location.rsplit("/", 1)[-1]
        if entity.id != id_returned_from_broker:
            raise NgsiApiError(
                f"Broker returned wrong id. Expected={entity.id} Returned={id_returned_from_broker}"
            )
        return entity

    @rfc7807_error_handle
    def get(
        self,
        eid: Union[EntityId, Entity],
        ctx: str = None,
        asdict: bool = False,
        **kwargs,
    ) -> Entity:
        eid = eid.id if isinstance(eid, Entity) else eid
        headers = {
            "Accept": "application/ld+json",
            "Content-Type": None,
        }  # overrides session headers
        if ctx is not None:
            headers[
                "Link"
            ] = f'<{ctx}>; rel="{JSONLD_CONTEXT}"; type="application/ld+json"'
        r = self._session.get(f"{self.url}/{eid}", headers=headers, **kwargs)
        self._client.raise_for_status(r)
        return r.json() if asdict else Entity.from_dict(r.json())

    @rfc7807_error_handle
    def delete(self, eid: Union[EntityId, Entity]) -> bool:
        eid = eid.id if isinstance(eid, Entity) else eid
        logger.info(f"{eid=}")
        logger.info(f"url={self.url}/{eid}")
        r = self._session.delete(f"{self.url}/{eid}")
        logger.info(f"requests: {r.request.url}")
        self._client.raise_for_status(r)
        return bool(r)

    @rfc7807_error_handle
    def exists(self, eid: Union[EntityId, Entity]) -> bool:
        eid = eid.id if isinstance(eid, Entity) else eid
        r = self._session.get(f"{self.url}/{eid}")
        if r:
            payload = r.json()
            return "@context" in payload
        return False

    @rfc7807_error_handle
    def upsert(self, entity: Entity) -> Entity:
        try:
            return self.create(entity)
        except NgsiAlreadyExistsError:
            self.delete(entity)
            return self.create(entity)

    @rfc7807_error_handle
    def update(self, entity: Entity, check_exists: bool = True) -> Optional[Entity]:
        if check_exists and self.exists(entity):
            self.delete(entity)
            return self.create(entity)
        return None

    @rfc7807_error_handle
    def query(
        self,
        type: str = None,
        q: str = None,
        ctx: str = None,
        limit: int = 0,
        offset: int = 0,
        **kwargs,
    ) -> List[Entity]:
        params = {}
        if limit != 0:
            params |= {"limit": limit}
        if offset != 0:
            params |= {"offset": offset}
        if type is None and q is None:
            raise ValueError("Must indicate at least a type or a query string")
        if type:
            params["type"] = type
        if q:
            params["q"] = q
        headers = {
            "Accept": "application/ld+json",
            "Content-Type": None,
        }  # overrides session headers
        if ctx is not None:
            headers[
                "Link"
            ] = f'<{ctx}>; rel="{JSONLD_CONTEXT}"; type="application/ld+json"'
        r = self._session.get(
            self.url,
            headers=headers,
            params=params,
        )
        r.raise_for_status()
        self._client.raise_for_status(r)
        entities = r.json()
        logger.debug(f"{entities=}")
        return [Entity.from_dict(entity) for entity in entities]

    @rfc7807_error_handle
    def count(self, type: str = None, q: str = None, **kwargs) -> int:
        params = {"limit": 0, "count": "true"}
        if type is None and q is None:
            raise ValueError("Must indicate at least a type or a query string")
        if type:
            params["type"] = type
        if q:
            params["q"] = q
        r = self._session.get(
            self.url,
            headers={
                "Accept": "application/json",
                "Content-Type": None,
            },  # overrides session headers
            params=params,
        )
        r.raise_for_status()
        self._client.raise_for_status(r)
        count = int(r.headers["NGSILD-Results-Count"])
        return count



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

from __future__ import annotations
from typing import TYPE_CHECKING, Optional
import re
import logging

if TYPE_CHECKING:
    from .client import Client

from ngsildclient.model.constants import CORE_CONTEXT
from .exceptions import rfc7807_error_handle


logger = logging.getLogger(__name__)


class Contexts:
    def __init__(self, client: Client, url: str):
        self._client = client
        self._session = client.session
        self.url = url

    @rfc7807_error_handle
    def list(self, pattern: str = None) -> Optional[dict]:
        r = self._session.get(f"{self.url}")
        contexts = r.json()
        if pattern is not None:
            contexts = [x for x in contexts if re.search(pattern, x, re.IGNORECASE)]
        return contexts

    @rfc7807_error_handle
    def get(self, ctx: str) -> dict:
        r = self._session.get(f"{self.url}/{ctx}")
        self._client.raise_for_status(r)
        return r.json()

    @rfc7807_error_handle
    def _delete(self, ctx: str) -> bool:
        r = self._session.delete(f"{self.url}/{ctx}")
        self._client.raise_for_status(r)
        return bool(r)

    @rfc7807_error_handle
    def delete(self, ctx: str, pattern: str = None) -> bool:
        if pattern is None:
            return self._delete(ctx)
        deleted = False
        for ctx in self.list():
            if re.search(pattern, ctx, re.IGNORECASE):
                deleted |= self.delete(ctx)
        return deleted

    @rfc7807_error_handle
    def exists(self, ctx: str) -> bool:
        r = self._session.get(f"{self.url}/{ctx}")
        if r:
            payload = r.json()
            return "@context" in payload
        return False

    @rfc7807_error_handle
    def cleanup(self) -> None:
        """Delete all contexts except the Core context."""
        for ctx in self.list():
            if ctx != CORE_CONTEXT:
                self.delete(ctx)

    @rfc7807_error_handle
    def add(self, ctx: dict) -> bool:
        if not ctx.get("@context"):
            raise ValueError("Expect a JSON object that has a top-level field named @context.")
        r = self._session.post(f"{self.url}/", json=ctx)
        self._client.raise_for_status(r)



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

"""A module that covers NGSI-LD API exceptions.

When possible, Exceptions are mapped from the API Error Types,
and enriched thanks to the ProblemDetails added by the API.
"""

import logging

from dataclasses import dataclass

from requests.exceptions import HTTPError, ContentDecodingError, RequestException
from requests import Response
from ..exceptions import NgsiError

logger = logging.getLogger(__name__)


@dataclass
class ProblemDetails:
    type: str
    title: str
    status: int
    detail: str
    instance: str = None
    extension: dict = None


class NgsiApiError(NgsiError):
    pass


class NgsiNotConnectedError(NgsiApiError):
    pass


class NgsiClientTooManyResultsError(NgsiApiError):
    pass


class NgsiHttpError(NgsiApiError):
    def __init__(self, statuscode: int):
        self.statuscode = statuscode
        self.message = f"HTTP {statuscode} error. No Problem Detail provided."
        super().__init__(self.message)


class NgsiContextBrokerError(NgsiApiError):
    def __init__(self, problemdetails: ProblemDetails):
        self.problemdetails = problemdetails
        super().__init__(f"{problemdetails.title} : {problemdetails.detail}")


class NgsiInvalidRequestError(NgsiContextBrokerError):
    pass


class NgsiBadRequestDataError(NgsiContextBrokerError):
    pass


class NgsiAlreadyExistsError(NgsiContextBrokerError):
    pass


class NgsiOperationNotSupportedError(NgsiContextBrokerError):
    pass


class NgsiResourceNotFoundError(NgsiContextBrokerError):
    pass


class NgsiInternalError(NgsiContextBrokerError):
    pass


class NgsiLdContextNotAvailableError(NgsiContextBrokerError):
    pass


class NgsiTooComplexQueryError(NgsiContextBrokerError):
    pass


class NgsiTooManyResultsError(NgsiContextBrokerError):
    pass


class NgsiNoMultiTenantSupportError(NgsiContextBrokerError):
    pass


class NgsiNonexistentTenantError(NgsiContextBrokerError):
    pass


ERRORTYPES = {
    "https://uri.etsi.org/ngsi-ld/errors/InvalidRequest": NgsiInvalidRequestError,
    "https://uri.etsi.org/ngsi-ld/errors/BadRequestData": NgsiBadRequestDataError,
    "https://uri.etsi.org/ngsi-ld/errors/AlreadyExists": NgsiAlreadyExistsError,
    "https://uri.etsi.org/ngsi-ld/errors/OperationNotSupported": NgsiOperationNotSupportedError,
    "https://uri.etsi.org/ngsi-ld/errors/ResourceNotFound": NgsiResourceNotFoundError,
    "https://uri.etsi.org/ngsi-ld/errors/InternalError": NgsiInternalError,
    "https://uri.etsi.org/ngsi-ld/errors/TooComplexQuery": NgsiTooComplexQueryError,
    "https://uri.etsi.org/ngsi-ld/errors/TooManyResults": NgsiTooManyResultsError,
    "https://uri.etsi.org/ngsi-ld/errors/LdContextNotAvailable": NgsiLdContextNotAvailableError,
    "https://uri.etsi.org/ngsi-ld/errors/NoMultiTenantSupport": NgsiNoMultiTenantSupportError,
    "https://uri.etsi.org/ngsi-ld/errors/NonexistentTenant": NgsiNonexistentTenantError,
}


def rfc7807_error_handle(func):
    """A decorator function to handle enriched Exceptions that accept a ProblemDetails instance.

    See Also
    --------
    api.entities.create
    api.entities.retrieve
    """

    def inner_function(*args, **kwargs):
        problemdetails: dict = {}
        try:
            return func(*args, **kwargs)
        except HTTPError as e:
            r: Response = e.response
            try:
                problemdetails = r.json()
                logger.info(f"{problemdetails=}")
            except ContentDecodingError:
                raise NgsiHttpError(r.status_code) from e
            try:
                pd_type = problemdetails.pop("type").rstrip()
                logger.info(f"{pd_type=}")
                exception: NgsiContextBrokerError = ERRORTYPES.get(pd_type)
                logger.info(f"{exception=}")
                pd = ProblemDetails(
                    pd_type,
                    problemdetails.pop("title", None),
                    r.status_code,
                    problemdetails.pop("detail", None),
                    problemdetails.pop("instance", None),
                    problemdetails,  # extension
                )
                raise exception(pd)
            except HTTPError as e:
                raise NgsiApiError(f"Error while requesting the broker API. Status code = {r.status_code}") from e
        except RequestException as e:
            raise NgsiApiError("Error while requesting the broker API") from e

    return inner_function



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

from ..exceptions import NgsiError


class NgsiModelError(NgsiError):
    pass


class NgsiMissingIdError(NgsiModelError):
    pass


class NgsiMissingTypeError(NgsiModelError):
    pass


class NgsiMissingContextError(NgsiModelError):
    pass


class NgsiUnmatchedAttributeTypeError(NgsiModelError):
    pass


class NgsiDateFormatError(NgsiModelError):
    pass


class NgsiJsonError(NgsiModelError):
    pass


class NgsiUnsupportedFormatError(NgsiModelError):
    pass



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

"""
This module contains a few helper functions to deal with UUIDs.
"""

import uuid
from base64 import b64encode


def uuidshortener(uuid: uuid.UUID) -> str:
    """Returns a "short" string representation of the UUID object.

    The string will be 22 characters long, base64-encoded, with padding characters removed.
    The encoding uses the urlsafe alphabet with a slightly difference.
    The dash character (sometimes used as a field separator) is replaced by the tilde character.

    Parameters
    ----------
    uuid : uuid.UUID
        The input UUID object

    Returns
    -------
    str
        Its string representation, 22 characters

    Example:
    --------
    >>> from uuid import uuid4
    >>> from ngsildclient.utils.uuid import uuidshortener
    >>> uuid = uuid4()
    >>> print(uuid)
    632dd95d-55bf-4f4e-9dd1-05f02531756f
    >>> shortid: str = uuidshortener(uuid)
    >>> print(shortid)
    Yy3ZXVW_T06d0QXwJTF1bw
    """
    # like in urlsafe_base64encode() but ~ replaces -
    return b64encode(uuid.bytes, altchars=b"~_").decode().rstrip("=")


def shortuuid(random: bool = False) -> str:
    """Returns a unique identifier, 22 characters long.

    May be useful in some cases to create a unique Entity identifier.
    The string will be 22 characters long, base64-encoded, with padding characters removed.
    The encoding uses the urlsafe alphabet with a slightly difference.
    The dash character (sometimes used as a field separator) is replaced by the tilde character.

    Parameters
    ----------
    random : bool, optional
        if set uses UUID1 else UUID4, by default False

    Returns
    -------
    str
        A short unique identifier, 22 characters longs.

    Example:
    --------
    >>> from ngsildclient import *
    >>> crop = Entity("AgriCrop", shortuuid())
    >>> crop.pprint()
    {
      "@context": [
          "https://uri.etsi.org/ngsi-ld/v1/ngsi-ld-core-context.jsonld"
      ],
      "id": "urn:ngsi-ld:AgriCrop:SQiKZZRYRnOYgeojIVz5lA",
      "type": "AgriCrop"
    }
    """
    uid = uuid.uuid1() if random else uuid.uuid4()
    return uuidshortener(uid)



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.


"""
This module contains helper functions to generate iso8601-formatted strings from native Python date types.

Notes
-----
NGSI-LD dates are expressed in UTC, using the ISO8601 format.

References
----------
.. [1] ETSI, 2021. "Supported data types for Values" in Context Information Management (CIM); NGSI-LD API
    ETSI GS CIM 009 V1.4.2, pp. 41-42, 2021-04.
"""

import re

from typing import Union, Optional
from datetime import datetime, date, time, timezone
from contextlib import suppress

from ngsildclient.model.constants import TemporalType

ISO8601_PATTERN = re.compile(r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z")


def from_datetime(value: datetime) -> str:
    """Convert from datetime to ISO8601-formatted string.

    Parameters
    ----------
    value : datetime
        The datetime to be converted. Expected in UTC.

    Returns
    -------
    str
        ISO8601-formatted string

    Examples
    --------
    >>> from datetime import datetime
    >>> from ngsildclient import iso8601
    >>> d = datetime(2021, 10, 13, 9, 29)
    >>> print(iso8601.from_datetime(d))
    2021-10-13T09:29:00Z
    """
    # convert to UTC if naive-datetime or not-UTC aware-datetime
    if value.tzinfo is None or value.tzinfo != timezone.utc:
        value = value.astimezone(timezone.utc)
    return value.strftime("%Y-%m-%dT%H:%M:%SZ")


def utcnow() -> str:
    """Converts the current UTC datetime to ISO8601-formatted string.

    Returns
    -------
    str
        The current UTC datetime, ISO8601-formatted
    """
    return datetime.now(timezone.utc)


def from_date(value: date) -> str:
    """Convert from date to ISO8601-formatted string.

    Parameters
    ----------
    value : date
        The date to be converted.

    Returns
    -------
    str
        ISO8601-formatted string

    Examples
    --------
    >>> from datetime import date
    >>> from ngsildclient import iso8601
    >>> d = date(2021, 10, 13)
    >>> print(iso8601.from_date(d))
    2021-10-13
    """
    return value.strftime("%Y-%m-%d")


def from_time(value: time) -> str:
    """Convert from time to ISO8601-formatted string.

    Parameters
    ----------
    value : time
        The time to be converted. Expected in UTC.

    Returns
    -------
    str
        ISO8601-formatted string

    Examples
    --------
    >>> from datetime import time
    >>> from ngsildclient import iso8601
    >>> d = time(9, 29)
    >>> print(iso8601.from_time(d))
    09:29:00Z
    """
    return value.strftime("%H:%M:%SZ")


def _from_string(value: str) -> tuple[str, TemporalType, datetime]:
    """Guess the temporal date type from a given string.

    This function should not be called by the end user. It is used internally by the `parse()` function.

    Parameters
    ----------
    value : str
        A string representation of either a datetime, date or time

    Returns
    -------
    tuple[str, TemporalType]
        A tuple composed of the `value` argument and the TemporalType that has been identified

    Raises
    ------
    ValueError
        The date format does not match datetime, date or time

    Examples
    --------
    >>> from ngsildclient import iso8601
    >>> print(iso8601._from_string("2021-10-13"))
    ('2021-10-13', <TemporalType.DATE: 'Date'>)
    """
    with suppress(ValueError):
        if len(value) == 20:
            dt = datetime.strptime(value, "%Y-%m-%dT%H:%M:%SZ")
            return value, TemporalType.DATETIME, dt
        elif len(value) == 10:
            datetime.strptime(value, "%Y-%m-%d")
            return value, TemporalType.DATE, None
        elif len(value) == 9:
            datetime.strptime(value, "%H:%M:%SZ")
            return value, TemporalType.TIME, None
    raise ValueError(f"Bad date format : {value}")


def parse(
    value: Union[datetime, date, time, str]
) -> tuple[str, TemporalType, datetime]:
    """Guess the temporal date type from a given argument carrying a temporal information.

    This function is typically used to build a NGSI-LD Temporal Property or temporal metadata such as `observedAt`.

    Parameters
    ----------
    value : Union[datetime, date, time, str]
        A value carrying temporal information, could be either either a datetime, date or time,
        or a string representation of these former types.

    Returns
    -------
    tuple[str, TemporalType]
        A tuple composed of a ISO8601 string representation of the date and the TemporalType that has been identified

    Raises
    ------
    ValueError
        The date format does not match datetime, date or time or a valid string representation of a date

    See Also
    --------
    ngsildclient.attribute

    Examples
    --------
    >>> from ngsildclient import iso8601
    >>> print(iso8601.parse(date(2021,9,13)))
    ('2021-10-13', <TemporalType.DATE: 'Date'>)
    """
    if isinstance(value, datetime):
        return from_datetime(value), TemporalType.DATETIME, value
    if isinstance(value, date):
        return from_date(value), TemporalType.DATE, None
    if isinstance(value, time):
        return from_time(value), TemporalType.TIME, None
    if isinstance(value, str):
        return _from_string(value)
    raise ValueError(f"Bad date format : {value}")


def extract(value: str) -> Optional[datetime]:
    """Extract an ISO8601 datetime string from the input string.

    Parameters
    ----------
    value : str
        the input string

    Returns
    -------
    datetime
        the extracted datetime if found, else None

    Examples
    --------
    >>> from ngsildclient.utils import iso8601
    >>> iso8601.extract("urn:ngsi-ld:WeatherObserved:Spain-WeatherObserved-Valladolid-2016-11-30T07:00:00Z")
    datetime.datetime(2016, 11, 30, 7, 0)
    """
    dates = ISO8601_PATTERN.findall(value)
    if len(dates) < 1:
        return None
    try:
        return datetime.strptime(dates[-1], "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=timezone.utc)
    except ValueError:
        return None



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

import logging
import requests
from dataclasses import dataclass
from typing import Optional, Tuple, Generator, List, Union, overload
from math import ceil

from ..utils import is_interactive
from ..model.entity import Entity
from .constants import *
from .entities import Entities
from .batch import BatchOp
from .types import Types
from .contexts import Contexts
from .subscriptions import Subscriptions
from .exceptions import *

logger = logging.getLogger(__name__)

"""This module contains the definition of the Client class.
"""


@dataclass
class Broker:
    """Represent a NGSI-LD Context Broker."""

    vendor: Vendor = Vendor.UNKNOWN
    version: str = "N/A"


class Client:
    """An implementation of the NGSI-LD client API.

    It allows to connect to a NGSI-LD Context Broker, check the connection and
    try to identify the vendor.
    As for now it focuses on the /entities/{entityId} endpoint.
    Allowed operations are :
    - create(), update(), upsert()
    - get(), exists()
    - delete()

    Update() and upsert() operations are not atomic, as they
    aren't provided as-is by the API, but require chaining 2 API calls.

    When encountering errors the Client throws enriched
    Exceptions, as NGSI-LD API supports ProblemDetails [IETF RFC 7807].

    Example:
    --------
    >>> from ngsildclient import *

    >>> # Here we don't build our own NGSI-LD entity, but grab an example from SmartDatamodels
    >>> farm = Entity.load(SmartDatamodels.SmartAgri.Agrifood.AgriFarm)

    >>> # farm.pprint()

    >>> # Connect to the Context Broker and upsert the entity
    >>> # As we don't provide any arguments to the Client constructor
    >>> # It connects to localhost and default port (without any authentication)
    >>> with Client() as client:
    >>>     client.upsert(farm)
    """

    def __init__(
        self,
        hostname: str = "localhost",
        port: int = NGSILD_DEFAULT_PORT,
        secure: bool = False,
        useragent: str = UA,
        tenant: str = None,
        overwrite: bool = False,
        ignore_errors: bool = False,
        proxy: str = None,
    ):
        """Create a Client instance to interact with the Context Broker.

        The Client allows to retrieve or send entities
        (model.Entity instances) to the Context Broker.
        For example, one can retrieve an entity from the Context Broker,
        modify it (update/delete/add properties),
        and send it back to the Context Broker.

        Parameters
        ----------
        hostname : str, optional
            the hostname to connect to, by default "localhost"
        port : int, optional
            the TCP port to connect to, by default NGSILD_DEFAULT_PORT
        secure : bool, optional
            whether to use TLS, by default False
        useragent : str, optional
            the User Agent string sent in the HTTP headers, by default UA
        tenant : str, optional
            the tenant string in case you make use of multi-tenancy, by default None
        overwrite : bool, optional
            if set create() will behave like upsert(), by default False
        ignore_errors : bool, optional
            if set tests the connection at init time and raisesan exception if failed, by default False
        proxy : str, optional
            proxies all requests to the provided proxy string (for debugging purpose), by default None

        See Also
        --------
        api.model.Entity

        """
        self.hostname = hostname
        self.port = port
        self.secure = secure
        self.scheme = "https" if secure else "http"
        self.url = f"{self.scheme}://{hostname}:{port}"
        self.basepath = f"{self.url}/{NGSILD_PATH}"
        self.useragent = useragent
        self.tenant = tenant
        self.overwrite = overwrite
        self.ignore_errors = ignore_errors
        self.proxy = proxy

        self.session = requests.Session()
        self.session.headers = {
            "User-Agent": self.useragent,
            "Accept": "application/ld+json",
            "Content-Type": "application/ld+json",
        }
        if tenant is not None:
            self.session.headers["NGSILD-Tenant"] = tenant
        if proxy:
            self.session.proxies = {proxy}

        logger.info("Connecting client ...")

        self._entities = Entities(self, f"{self.url}/{ENDPOINT_ENTITIES}")
        self._batch = BatchOp(self, f"{self.url}/{ENDPOINT_BATCH}")
        self._types = Types(self, f"{self.url}/{ENDPOINT_TYPES}")
        self._contexts = Contexts(self, f"{self.url}/{ENDPOINT_CONTEXTS}")
        self._subscriptions = Subscriptions(self, f"{self.url}/{ENDPOINT_SUBSCRIPTIONS}")

        self.broker = Broker(Vendor.UNKNOWN, "N/A")

        # get status and retrieve Context Broker information
        status = self.is_connected(raise_for_disconnected=True)
        if status and is_interactive():
            self.broker = Broker(*self.guess_vendor())
            print(self._welcome_message())
        else:
            print(self._fail_message())

    def raise_for_status(self, r: Response):
        """Raises an exception depending on the API response.

        Parameters
        ----------
        r : Response
            Response from the Context Broker
        """
        if not self.ignore_errors:
            r.raise_for_status()

    def is_connected(self, raise_for_disconnected=False) -> bool:
        """Test if connection to Context Broker is established.

        Send a valid test request to the Context Broker and expects a result.

        Parameters
        ----------
        raise_for_disconnected : bool, optional
            throw an exception if connection fails (if not set only return False), by default False

        Returns
        -------
        bool
            True if the Context Broker replies

        Raises
        ------
        NgsiNotConnectedError
        """
        url = f"{self.url}/{ENDPOINT_ENTITIES}"
        params = {"type": "None", "limit": 0, "count": "true"}
        try:
            r = self.session.get(
                url,
                headers={
                    "Accept": "application/json",
                    "Content-Type": None,
                },  # overrides session headers
                params=params,
            )
            r.raise_for_status()
        except Exception as e:
            if is_interactive():
                return False
            if raise_for_disconnected:
                raise NgsiNotConnectedError(f"Cannot connect to Context Broker at {self.hostname}:{self.port}: {e}")
            else:
                logger.error(e)
                return False
        return True

    @property
    def version(self) -> str:
        return __version__

    @property
    def entities(self):
        return self._entities

    @property
    def batch(self):
        return self._batch

    @property
    def types(self):
        return self._types

    @property
    def contexts(self):
        return self._contexts

    @property
    def subscriptions(self):
        return self._subscriptions

    def close(self):
        """Terminates the client.

        Closes the underlying Requests.Session.
        """
        self.session.close()

    @overload
    def create(self, entity: Entity, skip: bool = False, overwrite: bool = False) -> Entity:
        """Create an entity.

        Facade method for Entities.create().

        Parameters
        ----------
        entity : Entity
            the entity to be created by the Context Broker
        skip : bool, optional
            if set, skips creation (do nothing) if already exists, by default False
        overwrite : bool, optional
            if set, force upsert the entity if already exists, by default False

        Returns
        -------
        Entity
            the entity succesfully created
        """
        ...

    @overload
    def create(self, entities: List[Entity], skip: bool = False, overwrite: bool = False):
        """Create a batch of entities.

        Facade method for Batch.create().

        Parameters
        ----------
        entities : List[Entity]
            the entity to be created by the Context Broker
        skip : bool, optional
            if set, skips creation (do nothing) if already exists, by default False
        overwrite : bool, optional
            if set, force upsert the entity if already exists, by default False

        Returns
        -------
        BatchOperationResult
            TODO
        """
        ...

    def create(
        self,
        _entities: Union[Entity, List[Entity]],
        skip: bool = False,
        overwrite: bool = False,
    ) -> Optional[Entity]:
        if isinstance(_entities, Entity):
            entity = _entities
            return self.entities.create(entity, skip, overwrite)
        else:
            return self.batch.create(_entities, skip, overwrite)

    def get(
        self,
        eid: Union[EntityId, Entity],
        ctx: str = None,
        asdict: bool = False,
        **kwargs,
    ) -> Entity:
        """Retrieve an entity given its id.

        Facade method for Entities.retrieve().
        If already dealing with an entity instance one can provide the entity itself instead of its id.

        Parameters
        ----------
        eid : Union[EntityId, Entity]
            The entity identifier or the entity instance
        ctx : str
            The context
        asdict : bool, optional
            If set (instead of returning an Entity) returns the raw API response (a Python dict that represents the JSON response), by default False

        Returns
        -------
        Entity
            The retrieved entity
        """
        return self.entities.get(eid, ctx, asdict, **kwargs)

    @overload
    def delete(self, eid: Union[EntityId, Entity]) -> bool:
        """Delete an entity given its id.

        Facade method for Entities.delete().
        If already dealing with an entity instance one can provide the entity itself instead of its id.

        Parameters
        ----------
        eid : Union[EntityId, Entity]
            The entity identifier or the entity instance

        Returns
        -------
        bool
            True if the entity has been succefully deleted
        """
        ...

    @overload
    def delete(self, eids: List[Union[EntityId, Entity]]) -> bool:
        """Delete entities given its id.

        Facade method for Batch.delete().
        If already dealing with entity instances one can provide the entities instead of ids.

        Parameters
        ----------
        eids : List[Union[EntityId, Entity]]
            The entities ids or instances

        Returns
        -------
        bool
            True if the entity has been succefully deleted
        """
        ...

    def delete(self, eids: Union[Union[EntityId, Entity], List[Union[EntityId, Entity]]]) -> bool:
        if isinstance(eids, list):
            return self.batch.delete(eids)
        else:
            eid = eids
            return self.entities.delete(eid)

    def delete_from_file(self, filename: str) -> Union[Entity, dict]:
        """Delete in the broker all entities present in the JSON file.

        Parameters
        ----------
        filename : str
            Points to the JSON input file that contains entities.
        """
        entities = Entity.load(filename)
        return self.delete(entities)

    def exists(self, eid: Union[EntityId, Entity]) -> bool:
        """Tests if an entity exists.

        Facade method for Entities.exists().
        If already dealing with an entity instance one can provide the entity itself instead of its id.

        Parameters
        ----------
        eid : Union[EntityId, Entity]
            The entity identifier or the entity instance

        Returns
        -------
        bool
            True if the entity exists
        """
        return self.entities.exists(eid)

    @overload
    def upsert(self, entity: Entity) -> Entity:
        """Upsert the entity or update it if already exists.

        Facade method for Entities.upsert().

        Parameters
        ----------
        entity : Entity
            The entity to be upserted by the Context Broker

        Returns
        -------
        Entity
            The entity successfully upserted
        """
        ...

    @overload
    def upsert(self, entities: List[Entity]) -> dict:
        """Upsert a batch of entities.

        Facade method for Batch.upsert().

        Parameters
        ----------
        entity : Entity
            The entity to be upserted by the Context Broker

        Returns
        -------
        Entity
            The entities successfully upserted
        """
        ...

    def upsert(self, entities: Union[Entity, List[Entity]]) -> Union[Entity, dict]:
        if isinstance(entities, Entity):
            entity = entities
            return self.entities.upsert(entity)
        else:
            return self.batch.upsert(entities)

    def bulk_import(self, filename: str) -> Union[Entity, dict]:
        """Upsert all entities from a JSON file.

        Parameters
        ----------
        filename : str
            Points to the JSON input file that contains entities.
        """
        entities = Entity.load(filename)
        return self.upsert(entities)

    @overload
    def update(self, entity: Entity) -> Optional[Entity]:
        """Update the entity.

        Facade method for Entities.update().

        Parameters
        ----------
        entity : Entity
            The entity to be updated by the Context Broker

        Returns
        -------
        Optional[Entity]
            The entity successfully updated (or None if not found)
        """
        return self.entities.update(entity)

    @overload
    def update(self, entities: List[Entity]) -> dict:
        """Update a batch of entities.

        Facade method for Batch.update().

        Parameters
        ----------
        entities : List[Entity]
            The entities to be updated by the Context Broker

        Returns
        -------
        Optional[Entity]
            The entity successfully updated (or None if not found)
        """
        ...

    def update(self, entities: Union[Entity, List[Entity]]) -> Union[Optional[Entity], dict]:
        if isinstance(entities, Entity):
            entity = entities
            return self.entities.update(entity)
        else:
            return self.batch.update(entities)

    def query_head(self, type: str = None, q: str = None, ctx: str = None, n: int = 5, **kwargs) -> List[Entity]:
        """Retrieve entities given its type and/or query string.

        Retrieve up to PAGINATION_LIMIT_MAX entities.
        Use query_all() to retrieve all entities.
        Use entities.query() to deal with limit and offset on your own.

        Parameters
        ----------
        etype : str
            The entity's type
        q: str
            The query string (NGSI-LD Query Language)
        ctx: str
            The context
        n: int
            The first n entities to be retrieved
        Returns
        -------
        list[Entity]
            Retrieved entities matching the given type and/or query string

        Example:
        --------
        >>> with Client() as client:
        >>>     client.query(type="AgriFarm") # match a given type

        >>> with Client() as client:
        >>>     client.query(type="AgriFarm", q='contactPoint[email]=="wheatfarm@email.com"') # match type and query
        """
        return self.entities.query(type, q, ctx, limit=n)

    def query_all(
        self,
        type: str = None,
        q: str = None,
        ctx: str = None,
        limit: int = PAGINATION_LIMIT_MAX,
        max: int = 1_000_000,
        **kwargs,
    ) -> List[Entity]:
        """Retrieve entities given its type and/or query string.

        Retrieve all entities by sending as many requests as needed, using pagination.
        Assume data hold in memory. Should not be an issue except for very large datasets.

        Parameters
        ----------
        etype : str
            The entity's type
        q: str
            The query string (NGSI-LD Query Language)
        ctx: str
            The context
        limit: int
            The number of entities retrieved in each request

        Returns
        -------
        list[Entity]
            Retrieved entities matching the given type and/or query string

        Example:
        --------
        >>> with Client() as client:
        >>>     client.query_all(type="AgriFarm") # match a given type

        >>> with Client() as client:
        >>>     client.query_all(type="AgriFarm", q='contactPoint[email]=="wheatfarm@email.com"') # match type and query
        """

        entities: list[Entity] = []
        count = self.entities.count(type, q)
        if count > max:
            raise NgsiClientTooManyResultsError(f"{count} results exceed maximum {max}")
        for page in range(ceil(count / limit)):
            entities.extend(self.entities.query(type, q, ctx, limit, page * limit))
        return entities

    def query_generator(
        self,
        type: str = None,
        q: str = None,
        ctx: str = None,
        limit: int = PAGINATION_LIMIT_MAX,
        batch: bool = False,
        **kwargs,
    ) -> Generator[Entity, None, None]:
        count = self.entities.count(type, q)
        for page in range(ceil(count / limit)):
            if batch:
                yield self.entities.query(type, q, ctx, limit, page * limit)
            else:
                yield from self.entities.query(type, q, ctx, limit, page * limit)

    def count(self, type: str = None, q: str = None, **kwargs) -> int:
        """Return number of entities matching type and/or query string.

        Facade method for Entities.count().

        Parameters
        ----------
        etype : str
            The entity's type
        query: str
            The query string (NGSI-LD Query Language)

        Returns
        -------
        int
            The number of matching entities

        Example:
        --------
        >>> with Client() as client:
        >>>     client.count(type="AgriFarm") # match a given type

        >>> with Client() as client:
        >>>     client.count(type="AgriFarm", query='contactPoint[email]=="wheatfarm@email.com"') # match type and query
        """
        return self.entities.count(type, q)

    def delete_where(self, type: str = None, q: str = None, **kwargs):
        """Batch delete entities matching type and/or query string.

        Parameters
        ----------
        etype : str
            The entity's type
        query: str
            The query string (NGSI-LD Query Language)

        Example:
        --------
        >>> with Client() as client:
        >>>     client.delete_where(type="AgriFarm", query='contactPoint[email]=="wheatfarm@email.com"') # match type and query
        """
        g = self.query_generator(type, q, batch=True, **kwargs)
        for batch in g:
            self.batch.delete(batch)

    def drop(self, type: str) -> None:
        """Batch delete entities matching the given type.

        Parameters
        ----------
        type : str
            The entity's type

        Example:
        --------
        >>> with Client() as client:
        >>>     client.drop("AgriFarm")
        """
        self.delete_where(type=type)

    def list_types(self) -> Optional[dict]:
        return self.types.list()

    def purge(self) -> None:
        """Batch delete all entities.

        Example:
        --------
        >>> with Client() as client:
        >>>     client.purge()
        """
        for type in self.types.list():
            self.drop(type)

    def flush_all(self) -> None:
        """Batch delete all entities and remove all contexts.

        Example:
        --------
        >>> with Client() as client:
        >>>     client.purge()
        """
        self.purge(type)
        self.contexts.cleanup()

    def guess_vendor(self) -> tuple[Vendor, Version]:
        """Try to guess the Context Broker vendor.

        According to its own API, by using version or status endpoint.

        Returns
        -------
        tuple[Vendor, Version]
            A tuple composed of the Vendor (if identified) and the version.

        Example:
        --------
        >>> with Client() as client:
        >>>     print(client.guess_vendor())
        (<Vendor.ORIONLD: 'Orion-LD'>, 'post-v0.8.1')
        """
        if broker := self._broker_version_orionld():
            return broker
        if broker := self._broker_version_java_spring():
            return broker
        return Vendor.UNKNOWN, "N/A"

    def _broker_version_orionld(self) -> Optional[str]:
        """Requests the broker looking for Orion-LD version.

        Targets the /version endpoint.

        Returns
        -------
        Optional[str]
            The Orion-LD version if found
        """
        url = f"{self.url}/version"
        headers = {
            "Accept": "application/json",
            "Content-Type": None,
        }  # overrides session headers
        try:
            r = self.session.get(url, headers=headers)
            r.raise_for_status()
            return Vendor.ORIONLD, r.json()["orionld version"]
        except Exception:
            return None

    def _broker_version_java_spring(self) -> Optional[Tuple[Vendor, str]]:
        """Requests the Java-Spring based broker looking for Vendor and Version.

        Targets the /actuator/info endpoint.

        Returns
        -------
        Optional[Tuple[Vendor, str]]
            A tuple composed of the Vendor and the broker version
        """
        url = f"{self.url}/actuator"
        headers = {
            "Accept": "application/json",
            "Content-Type": None,
        }  # overrides session headers
        try:
            r = self.session.get(f"{url}/health", headers=headers)
            r.raise_for_status()
        except Exception:
            return None
        try:
            r = self.session.get(f"{url}/info", headers=headers)
            r.raise_for_status()
            build = r.json()["build"]
            version = build["version"]
            group = build["group"]
            if group == "eu.neclab.ngsildbroker":
                vendor = Vendor.SCORPIO
            elif group == "com.egm.stellio":
                vendor = Vendor.STELLIO
            else:
                return None
            return vendor, version
        except Exception:
            if is_interactive():
                print("Java-Spring based Context Broker detected. Try to enable info endpoint.")
            return None

    def _welcome_message(self) -> str:
        return f"Connected to Context Broker at {self.hostname}:{self.port} | vendor={self.broker.vendor.value} | version={self.broker.version}"

    def _fail_message(self) -> str:
        return f"Failed to connect to Context Broker at {self.hostname}:{self.port}"

    def _warn_spring_message(self) -> str:
        return "Java-Spring based Context Broker detected. Info endpoint disabled."

    # below the context manager methods

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

"""
This module contains a few helper functions to deal with URLs.
"""

import urllib.parse
import re


URL_PATTERN = re.compile(r"^http[s]{0,1}://")
"""Simple regex pattern that matches on strings starting with URL scheme (`regex.Pattern`).
"""


def escape(value: str) -> str:
    """URLEncode an URL.

    Parameters
    ----------
    value : str
        String representation of the URL

    Returns
    -------
    str
        The encoded URL as a string

    Examples
    --------
    >>> from ngsildclient.utils import url
    >>> print(url.escape("https://example.com?query=dummy&limit=5"))
    https%3A//example.com%3Fquery%3Ddummy%26limit%3D5

    See Also
    --------
    ngsildclient.utils.url.unescape
    """
    return urllib.parse.quote(value)


def unescape(value: str) -> str:
    """URLDecode an URL.

    Parameters
    ----------
    value : str
        String representation of the encoded URL

    Returns
    -------
    str
        The encoded URL as a string

    Examples
    --------
    >>> from ngsildclient.utils import url
    >>> print(url.escape("https%3A//example.com%3Fquery%3Ddummy%26limit%3D5"))
    https://example.com?query=dummy&limit=5

    See Also
    --------
    ngsildclient.utils.url.escape
    """
    return urllib.parse.unquote(value)


def isurl(value: str) -> bool:
    """Check if the given string represents an URL.

    Just test whether it starts with "http://" or "https://"

    Parameters
    ----------
    value : str
        The given string to be checked

    Returns
    -------
    bool
        True if the string looks like an URL

    Examples
    --------
    >>> from ngsildclient.utils import url
    >>> print(url.isurl("https://example.com?query=dummy&limit=5"))
    True
    """
    return URL_PATTERN.match(value) is not None



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

"""Contains helper modules to handle iso8601 dates, URLs, NGSI-LD URNs, and generate short uuids.
"""

import sys


def is_interactive() -> bool:
    return hasattr(sys, "ps1") or sys.flags.interactive



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

from __future__ import annotations
from typing import Union
from dataclasses import dataclass
from datetime import time
from enum import Enum

TimeOrStr = Union[time, str]  # type alias


class WeekDay(Enum):
    MONDAY = "Monday"
    TUESDAY = "Tuesday"
    WEDNESDAY = "Wednesday"
    THURSDAY = "Thursday"
    FRIDAY = "Friday"
    SATURDAY = "Saturday"
    SUNDAY = "Sunday"


WEEK = [day for day in WeekDay]
WEEK_END = [WeekDay.SATURDAY, WeekDay.SUNDAY]
WORKING_DAYS = [day for day in WEEK if day not in WEEK_END]


@dataclass
class OpeningHoursSpecification:
    """An OpeningHoursSpecification as described here : https://schema.org/OpeningHoursSpecification

    Simplified. Support a single open timeslot per day.
    """

    opens: str
    closes: str
    dayofweek: str

    def to_dict(self):
        return {"opens": self.opens, "closes": self.closes, "dayOfWeek": self.dayofweek}


class OpeningHoursBuilder:
    """A helper class that allows to easily build an openingHours property.

    Simplified. Support a single open timeslot per day.

    Example:
    --------
    >>> from ngsildclient import *
    >>> builder = OpeningHoursBuilder()
    >>> openinghours = builder.businessdays("10:00", "17:30").saturday("10:00", "14:00").build()
    >>> # Add an openingHours property to the entity you're creating
    >>> library = Entity("Library", "MyLibrary")
    >>> library.prop("openingHours", openinghours)
    >>> library.pprint()
    {
        "@context": [
            "https://uri.etsi.org/ngsi-ld/v1/ngsi-ld-core-context.jsonld"
        ],
        "id": "urn:ngsi-ld:Library:MyLibrary",
        "type": "Library",
        "openingHours": {
            "type": "Property",
            "value": [
            {
                "opens": "10:00",
                "closes": "17:30",
                "dayOfWeek": "Monday"
            },
            {
                "opens": "10:00",
                "closes": "17:30",
                "dayOfWeek": "Tuesday"
            },
            {
                "opens": "10:00",
                "closes": "17:30",
                "dayOfWeek": "Wednesday"
            },
            {
                "opens": "10:00",
                "closes": "17:30",
                "dayOfWeek": "Thursday"
            },
            {
                "opens": "10:00",
                "closes": "17:30",
                "dayOfWeek": "Friday"
            },
            {
                "opens": "10:00",
                "closes": "14:00",
                "dayOfWeek": "Saturday"
            }
            ]
        }
    }
    """

    def __init__(self):
        self._oh: dict = {}

    @staticmethod
    def _converttimes(opens, closes) -> tuple[str, str]:
        opens = opens if isinstance(opens, str) else opens.strftime("%H:%M")
        closes = closes if isinstance(closes, str) else closes.strftime("%H:%M")
        return opens, closes

    def monday(self, opens: TimeOrStr, closes: TimeOrStr):
        opens, closes = OpeningHoursBuilder._converttimes(opens, closes)
        self._oh[WeekDay.MONDAY] = OpeningHoursSpecification(
            opens, closes, WeekDay.MONDAY.value
        )
        return self

    def tuesday(self, opens: TimeOrStr, closes: TimeOrStr):
        opens, closes = OpeningHoursBuilder._converttimes(opens, closes)
        self._oh[WeekDay.TUESDAY] = OpeningHoursSpecification(
            opens, closes, WeekDay.TUESDAY.value
        )
        return self

    def wednesday(self, opens: TimeOrStr, closes: TimeOrStr):
        opens, closes = OpeningHoursBuilder._converttimes(opens, closes)
        self._oh[WeekDay.WEDNESDAY] = OpeningHoursSpecification(
            opens, closes, WeekDay.WEDNESDAY.value
        )
        return self

    def thursday(self, opens: TimeOrStr, closes: TimeOrStr):
        opens, closes = OpeningHoursBuilder._converttimes(opens, closes)
        self._oh[WeekDay.THURSDAY] = OpeningHoursSpecification(
            opens, closes, WeekDay.THURSDAY.value
        )
        return self

    def friday(self, opens: TimeOrStr, closes: TimeOrStr):
        opens, closes = OpeningHoursBuilder._converttimes(opens, closes)
        self._oh[WeekDay.FRIDAY] = OpeningHoursSpecification(
            opens, closes, WeekDay.FRIDAY.value
        )
        return self

    def saturday(self, opens: TimeOrStr, closes: TimeOrStr):
        opens, closes = OpeningHoursBuilder._converttimes(opens, closes)
        self._oh[WeekDay.SATURDAY] = OpeningHoursSpecification(
            opens, closes, WeekDay.SATURDAY.value
        )
        return self

    def sunday(self, opens: TimeOrStr, closes: TimeOrStr):
        opens, closes = OpeningHoursBuilder._converttimes(opens, closes)
        self._oh[WeekDay.SUNDAY] = OpeningHoursSpecification(
            opens, closes, WeekDay.SUNDAY.value
        )
        return self

    def days(self, opens: TimeOrStr, closes: TimeOrStr, *days):
        opens, closes = OpeningHoursBuilder._converttimes(opens, closes)
        for day in days:
            self._oh[day] = OpeningHoursSpecification(opens, closes, day.value)
        return self

    def weekend(self, opens: TimeOrStr, closes: TimeOrStr):
        return self.days(opens, closes, *WEEK_END)

    def businessdays(self, opens: TimeOrStr, closes: TimeOrStr, *exceptdays):
        openingdays = [day for day in WORKING_DAYS if day not in exceptdays]
        return self.days(opens, closes, *openingdays)

    def wholeweek(self, opens: TimeOrStr, closes: TimeOrStr, *exceptdays):
        return self.days(opens, closes, *WEEK)

    def build(self) -> dict:
        openingdays = []
        for day in WeekDay:
            if oh := self._oh.get(day, None):
                openingdays.append(oh.to_dict())
        return openingdays



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

from dataclasses import dataclass
from ...model.constants import CORE_CONTEXT


@dataclass
class NotificationParams:
    uri: str
    attrs: list[str] = None
    format: str = "normalized"

    def to_dict(self) -> dict:
        d = {}
        if self.attrs:
            d["attributes"] = self.attrs
        d["format"] = self.format
        d["endpoint"] = {"uri": self.uri, "accept": "application/ld+json"}
        return d


@dataclass
class Subscription:
    notification: NotificationParams
    id: str = None
    type: str = "Subscription"
    name: str = None
    description: str = None
    entities: list[dict] = None  # id, idPattern, or type
    watched_attrs: list[str] = None
    # time_interval: int = 0
    query: str = None
    # gquery: str = None
    # csf: str = None
    active: bool = True
    # expiresat: datetime = None
    # throttling: int = 0
    # tquery: str = None
    # scope: str = None
    # lang: str = None
    ctx: str = CORE_CONTEXT

    def to_dict(self) -> dict:
        d = {}
        if self.id:
            d["id"] = self.id
        d["type"] = self.type
        if self.name:
            d["name"] = self.name
        if self.description:
            d["description"] = self.description
        if self.entities:
            d["entities"] = self.entities
        if self.watched_attrs:
            d["watchedAttributes"] = self.watched_attrs
        if self.query:
            d["q"] = self.query
        d["isActive"] = self.active
        d["notification"] = self.notification.to_dict()
        d["@context"] = self.ctx
        return d


class SubscriptionBuilder:
    def __init__(
        self,
        uri: str,
    ):
        notification = NotificationParams(uri)
        self._subscr = Subscription(notification)
        self._subscr.entities = []

    def id(self, value: str):
        if not isinstance(value, str):
            raise ValueError("id shall be a string")
        self._subscr.id = value
        return self

    def name(self, value: str):
        if not isinstance(value, str):
            raise ValueError("name shall be a string")
        self._subscr.name = value
        return self

    def description(self, value: str):
        if not isinstance(value, str):
            raise ValueError("description shall be a string")
        self._subscr.description = value
        return self

    def select_id(self, value: str):
        if not isinstance(value, str):
            raise ValueError("EntitySelector id shall be a string")
        self._subscr.entities.append({"id": value})
        return self

    def select_idpattern(self, value: str):
        if not isinstance(value, str):
            raise ValueError("EntitySelector idPattern shall be a string")
        self._subscr.entities.append({"idPattern": value})
        return self

    def select_type(self, value: str):
        if not isinstance(value, str):
            raise ValueError("EntitySelector type shall be a string")
        self._subscr.entities.append({"type": value})
        return self

    def watch(self, value: list[str]):
        if not isinstance(value, list):
            raise ValueError("watchedAttributes shall be a list of strings")
        if value == []:
            raise ValueError("Empty array is not allowed")
        self._subscr.watched_attrs = value
        return self

    def query(self, value: str):
        if not isinstance(value, str):
            raise ValueError("query shall be a string")
        self._subscr.query = value
        return self

    def notif(self, value: list[str]):
        if not isinstance(value, list):
            raise ValueError("attribute names shall be a list of strings")
        if value == []:
            raise ValueError("Empty array is not allowed")
        self._subscr.notification.attrs = value
        return self

    def context(self, value: str):
        if not isinstance(value, str):
            raise ValueError("context shall be a string")
        self._subscr.ctx = value
        return self

    def build(self) -> dict:
        if self._subscr.entities == [] and self._subscr.watched_attrs is None:
            raise ValueError(
                "At least one of (a) entities or (b) watchedAttributes shall be present."
            )
        return self._subscr.to_dict()



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

from __future__ import annotations

from typing import Any, Union, List
from functools import reduce
from datetime import datetime
from geojson import Point, LineString, Polygon, MultiPoint

import ngsildclient.model.entity as entity
from ..utils import iso8601, url
from ..utils.urn import Urn
from .constants import *
from .exceptions import *

import json
import operator

"""This module contains the definition of the NgsiDict class.
"""


class NgsiDict(dict):
    """This class is a custom dictionary that backs an entity.

    NgsiDict is used to build and hold the entity properties, as well as the entity's root.
    It's not exposed to the user but intended to be used by the Entity class.
    NgsiDict provides methods that allow to build a dictionary compliant with a NGSI-LD structure.

    See Also
    --------
    model.Entity
    """

    def __init__(self, *args, dtcached: datetime = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._dtcached: datetime = dtcached if dtcached else iso8601.utcnow()

    @classmethod
    def _from_json(cls, payload: str):
        d = json.loads(payload)
        return cls(d)

    def _cachedate(self, dt: datetime):
        self._dtcached = dt

    def _dateauto(self):
        if self._dtcached is None:
            self._dtcached = iso8601.utcnow()
        return self._dtcached

    def __getitem__(self, element: str):
        return reduce(dict.__getitem__, element.split("."), self)

    def __delitem__(self, element: str):
        try:
            nested, k = element.rsplit(".", 1)
        except ValueError:
            dict.__delitem__(self, element)
        else:
            dict.__delitem__(self[nested], k)

    def __setitem__(self, key: str, value: Any):
        try:
            nested, k = key.rsplit(".", 1)
        except ValueError:
            dict.__setitem__(self, key, value)
        else:
            dict.__setitem__(self[nested], k, value)

    @classmethod
    def _load(cls, filename: str):
        with open(filename, "r") as fp:
            d = json.load(fp)
            return cls(d)

    def to_json(self, indent=None) -> str:
        """Returns the dict in json format"""
        return json.dumps(self, default=str, ensure_ascii=False, indent=indent)

    def pprint(self, *args, **kwargs) -> None:
        """Returns the dict pretty-json-formatted"""
        print(self.to_json(indent=2, *args, **kwargs))

    def _save(self, filename: str, indent=2):
        with open(filename, "w") as fp:
            json.dump(self, fp, default=str, ensure_ascii=False, indent=indent)

    def _process_observedat(self, observedat):
        if observedat is Auto:
            observedat = self._dateauto()
        date_str, temporaltype, dt = iso8601.parse(observedat)
        if temporaltype != TemporalType.DATETIME:
            raise NgsiDateFormatError(f"observedAt must be a DateTime : {date_str}")
        self._cachedate(dt)
        return date_str

    def _build_property(
        self,
        value: Any,
        unitcode: str = None,
        observedat: Union[str, datetime, type[Auto]] = None,
        datasetid: str = None,
        userdata: NgsiDict = None,
        escape: bool = False,
    ) -> NgsiDict:
        property: NgsiDict = NgsiDict()
        property["type"] = AttrType.PROP.value  # set type
        if isinstance(value, (int, float, bool, list, dict)):
            v = value
        elif isinstance(value, str):
            v = url.escape(value) if escape else value
        else:
            raise NgsiUnmatchedAttributeTypeError(f"Cannot map {type(value)} to NGSI type. {value=}")
        property["value"] = v  # set value
        if unitcode is not None:
            property[META_ATTR_UNITCODE] = unitcode
        if observedat is not None:
            property[META_ATTR_OBSERVED_AT] = self._process_observedat(observedat)
        if datasetid is not None:
            property[META_ATTR_DATASET_ID] = Urn.prefix(datasetid)
        if userdata:
            property |= userdata
        return property

    def prop(self, name: str, value: str, **kwargs):
        self[name] = self._build_property(value, **kwargs)
        return self[name]

    def _build_geoproperty(
        self,
        value: NgsiGeometry,
        observedat: Union[str, datetime, type[Auto]] = None,
        datasetid: str = None,
    ) -> NgsiDict:
        property: NgsiDict = NgsiDict()
        property["type"] = AttrType.GEO.value  # set type
        if isinstance(value, (Point, LineString, Polygon, MultiPoint)):
            geometry = value
        elif isinstance(value, tuple) and len(value) == 2:  # simple way for a location Point
            lat, lon = value
            geometry = Point((lon, lat))
        else:
            raise NgsiUnmatchedAttributeTypeError(f"Cannot map {type(value)} to NGSI type. {value=}")
        property["value"] = geometry  # set value
        if observedat is not None:
            property[META_ATTR_OBSERVED_AT] = self._process_observedat(observedat)
        if datasetid is not None:
            property[META_ATTR_DATASET_ID] = Urn.prefix(datasetid)
        return property

    def gprop(self, name: str, value: str, **kwargs):
        self[name] = self._build_geoproperty(value, **kwargs)
        return self[name]

    def _build_temporal_property(self, value: Union[NgsiDate, type[Auto]]) -> NgsiDict:
        property: NgsiDict = NgsiDict()
        property["type"] = AttrType.TEMPORAL.value  # set type

        if value is Auto:
            value = self._dateauto()

        date_str, temporaltype, dt = iso8601.parse(value)
        v = {
            "@type": temporaltype.value,
            "@value": date_str,
        }
        property["value"] = v  # set value
        self._cachedate(dt)
        return property

    def tprop(self, name: str, value: str, **kwargs):
        self[name] = self._build_temporal_property(value, **kwargs)
        return self[name]

    def _build_relationship(
        self,
        value: Union[str, List[str]],
        observedat: Union[str, datetime, type[Auto]] = None,
        datasetid: str = None,
        userdata: NgsiDict = None,
    ) -> NgsiDict:
        property: NgsiDict = NgsiDict()
        property["type"] = AttrType.REL.value  # set types
        property["object"] = Urn.prefix(value.id) if isinstance(value, entity.Entity) else Urn.prefix(value)
        if observedat is not None:
            property[META_ATTR_OBSERVED_AT] = self._process_observedat(observedat)
        if datasetid is not None:
            property[META_ATTR_DATASET_ID] = Urn.prefix(datasetid)
            if userdata:
                property |= userdata
        return property

    def _m_build_relationship(
        # Multiple Relationship limitation : no metadata
        self,
        value: List[str],
        observedat: Union[str, datetime, type[Auto]] = None,
        datasetid: str = None,
        userdata: NgsiDict = None,
    ) -> NgsiDict:
        property: List[NgsiDict] = []
        if isinstance(observedat, List):
            if len(observedat) != len(value):
                raise ValueError("Missing observedAt attributes")
        else:
            observedat = [observedat] * len(value)
        if isinstance(datasetid, List):
            if len(datasetid) != len(value):
                raise ValueError("Missing datasetId attributes")
        else:
            datasetid = [datasetid] * len(value)
        if isinstance(userdata, List):
            if len(userdata) != len(value):
                raise ValueError("Missing userdata attributes")
        else:
            userdata = [userdata] * len(value)
        for x in zip(value, observedat, datasetid, userdata):
            property.append(self._build_relationship(*x))
        return property

    def rel(self, name: str, value: str, **kwargs):
        if isinstance(name, Rel):
            name = name.value
        self[name] = self._build_relationship(value, **kwargs)
        return self[name]



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

from __future__ import annotations
from typing import TYPE_CHECKING, Optional

import logging

if TYPE_CHECKING:
    from .client import Client

from .exceptions import rfc7807_error_handle


logger = logging.getLogger(__name__)


class Types:
    def __init__(self, client: Client, url: str):
        self._client = client
        self._session = client.session
        self.url = url

    @rfc7807_error_handle
    def list(self) -> Optional[dict]:
        r = self._session.get(f"{self.url}")
        return r.json()["typeList"]



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

import http.client
import logging
import sys

__version__ = "0.1.10"

from .utils import iso8601, is_interactive
from .utils.uuid import shortuuid
from .model.entity import Entity
from .model.helper.postal import PostalAddressBuilder
from .model.helper.openinghours import OpeningHoursBuilder
from .model.constants import (
    CORE_CONTEXT,
    NESTED,
    Auto,
    SmartDataModels,
    Rel,
    TZ_UTC,
    TZ_WET,
    TZ_CET,
    TZ_FET,
)
from .model.mock import MockerNgsi
from .api.client import Client
from .api.helper.subscription import SubscriptionBuilder
from .exceptions import NgsiError
from .model.exceptions import NgsiModelError
from .api.exceptions import NgsiApiError, NgsiContextBrokerError, NgsiAlreadyExistsError


__all__ = [
    "iso8601",
    "shortuuid",
    "Entity",
    "PostalAddressBuilder",
    "OpeningHoursBuilder",
    "CORE_CONTEXT",
    "NESTED",
    "Auto",
    "Rel",
    "TZ_UTC",
    "TZ_WET",
    "TZ_CET",
    "TZ_FET",
    "MockerNgsi",
    "Client",
    "SubscriptionBuilder",
    "SmartDataModels",
    "NgsiError",
    "NgsiModelError",
    "NgsiApiError",
    "NgsiContextBrokerError",
    "NgsiAlreadyExistsError",
]

logging.basicConfig(
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=logging.INFO
)
if is_interactive():
    logging.disable(logging.CRITICAL)
    sys.tracebacklimit = 0

logger = logging.getLogger(__name__)

http.client.HTTPConnection.debuglevel = 1


def print_to_log(*args):
    logger.debug(" ".join(args))


# monkey patch the http.client's print() function
http.client.print = print_to_log



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

"""This module contains constants used in the api package.
"""

from ngsildclient import __version__

from typing import TYPE_CHECKING, Union, List
from enum import Enum, unique

if TYPE_CHECKING:
    from .entities import Entity

    OneOrManyEntities = Union[Entity, List[Entity]]

Version = str
EntityId = str


@unique
class Vendor(Enum):
    ORIONLD = "Orion-LD"
    SCORPIO = "Scorpio"
    STELLIO = "Stellio"
    CASSIOPEIA = "Cassiopeia"
    UNKNOWN = None


@unique
class AttrsFormat(Enum):
    NORMALIZED = "normalized"
    KEYVALUES = "keyValues"
    VALUES = "values"


@unique
class LogLevel(Enum):
    NONE = "NONE"
    FATAL = "FATAL"
    ERROR = "ERROR"
    WARN = "WARN"
    INFO = "INFO"
    DEBUG = "DEBUG"


UA = f"ngsildclient v{__version__}"

NGSILD_PATH = "ngsi-ld"
NGSILD_VERSION = "v1"
NGSILD_BASEPATH = f"{NGSILD_PATH}/{NGSILD_VERSION}"

NGSILD_DEFAULT_PORT_ORIONLD = 1026
NGSILD_DEFAULT_PORT_SCORPIO = 9090
NGSILD_DEFAULT_PORT_STELLIO = 8080
NGSILD_DEFAULT_PORT = NGSILD_DEFAULT_PORT_ORIONLD

# JSON-LD Vocabulary at w3.org
JSONLD_CONTEXT = "http://www.w3.org/ns/json-ld#context"

# endpoints MUST NOT end with a slash
ENDPOINT_STATUS = "version"
ENDPOINT_ADMIN = "admin"
ENDPOINT_LOG = f"{ENDPOINT_ADMIN}/log"
ENDPOINT_ENTITIES = f"{NGSILD_BASEPATH}/entities"
ENDPOINT_BATCH = f"{NGSILD_BASEPATH}/entityOperations"
ENDPOINT_TYPES = f"{NGSILD_BASEPATH}/types"
ENDPOINT_CONTEXTS = f"{NGSILD_BASEPATH}/jsonldContexts"
ENDPOINT_SUBSCRIPTIONS = f"{NGSILD_BASEPATH}/subscriptions"


PAGINATION_LIMIT_MAX = 100  # pagination

DEFAULT_ATTR_FORMAT = None  # Let Orion default value => AttrsFormat.NORMALIZED
DEFAULT_LOGLEVEL = LogLevel.WARN

WRITE_MODE_UPSERT = True
WRITE_MODE_IGNORE = False
DEFAULT_WRITE_MODE = WRITE_MODE_UPSERT

NOW = None



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

import logging
from typing import Callable
from ngsildclient import Entity
from ngsildclient.utils.uuid import shortuuid

logger = logging.getLogger(__name__)


def mock_id(e: Entity):
    mock_id = shortuuid()
    e.id = f"{e.id}:Mocked:{mock_id}"


def mock_payload(e: Entity):
    e.prop("mocked", True)


class MockerNgsi:
    def __init__(
        self,
        f_mock_id: Callable = mock_id,
        f_mock_payload: Callable = mock_payload,
    ):
        self.f_mock_id = f_mock_id
        self.f_mock_payload = f_mock_payload

    def _mock(self, entity: Entity) -> Entity:
        dst: Entity = entity.copy()
        self.f_mock_id(dst)
        self.f_mock_payload(dst)
        return dst

    def mock(self, entity: Entity, n: int = 10) -> list[Entity]:
        return [self._mock(entity) for _ in range(n)]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ngsildclient',
 'ngsildclient.api',
 'ngsildclient.api.helper',
 'ngsildclient.model',
 'ngsildclient.model.helper',
 'ngsildclient.utils']

package_data = \
{'': ['*']}

install_requires = \
['geojson>=2.5.0,<3.0.0', 'requests>=2.26.0,<3.0.0', 'rich>=12.4.1,<13.0.0']

setup_kwargs = {
    'name': 'ngsildclient',
    'version': '0.1.10',
    'description': 'A Python library that helps building NGSI-LD entities and interacting with a NGSI-LD Context Broker',
    'long_description': '# The ngsildclient library\n\n[![NGSI-LD badge](https://img.shields.io/badge/NGSI-LD-red.svg)](https://www.etsi.org/deliver/etsi_gs/CIM/001_099/009/01.02.01_60/gs_CIM009v010201p.pdf)\n[![SOF support badge](https://nexus.lab.fiware.org/repository/raw/public/badges/stackoverflow/fiware.svg)](http://stackoverflow.com/questions/tagged/fiware)\n<br>\n[![License](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](https://opensource.org/licenses/Apache-2.0)\n[![Read the Docs](https://img.shields.io/readthedocs/ngsildclient)](https://ngsildclient.readthedocs.io/en/latest/index.html)\n<br>\n[![deploy status](https://github.com/Orange-OpenSource/python-ngsild-client/workflows/CI/badge.svg)](https://github.com/Orange-OpenSource/python-ngsild-client/actions)\n[![PyPI](https://img.shields.io/pypi/v/ngsildclient.svg)](https://pypi.org/project/ngsildclient/)\n[![Python version](https://img.shields.io/pypi/pyversions/ngsildclient)](https://pypi.org/project/ngsildclient/)\n\n\n## Overview\n\n **ngsildclient** is a Python library that helps building NGSI-LD entities and allows to interact with a NGSI-LD Context Broker.\n\n The library primary purpose is to **ease and speed up the development of a NGSI Agent** and is also **useful for Data Modeling in the design stage**.\n\n## Key Features\n\n### Build NGSI-LD entities\n\nThe task of building a large NGSI-LD compliant entity is tedious, error-prone and results in a significant amount of code. \n\n**ngsildclient** provides primitives to build and manipulate NGSI-LD compliant entities without effort, in respect with the [ETSI specifications](https://www.etsi.org/committee/cim).\n\n### Implement the NGSI-LD API\n\n**ngsildclient** provides a NGSI-LD API Client implementation.\n\nActing as a Context Producer/Consumer **ngsildclient** is able to send/receive NGSI-LD entities to/from the Context Broker for creation and other operations.\n\nThe library wraps a large subset of the API endpoints and supports batch operations, queries, subscriptions.\n\n## Where to get it\n\nThe source code is currently hosted on GitHub at :\nhttps://github.com/Orange-OpenSource/python-ngsild-client\n\nBinary installer for the latest released version is available at the [Python\npackage index](https://pypi.org/project/ngsildclient).\n\n```sh\npip install ngsildclient\n```\n\n## Installation\n\n**ngsildclient** requires Python 3.9+.\n\nOne should use a virtual environment. For example with pyenv.\n\n```sh\nmkdir myagent && cd myagent\npyenv virtualenv 3.10.4 myagent\npyenv local myagent\npip install ngsildclient\n```\n\n## Getting started\n\nThe following code snippet builds a NGSI-LD entity related to a measure of air quality in Bordeaux then sends it to the Context Broker.\n\n```python\nfrom ngsildclient import Entity, Client\n\ne = Entity("AirQualityObserved", "Bordeaux-AirProbe42-2022-03-24T09:00:00Z")\ne.tprop("dateObserved").gprop("location", (44.84044, -0.5805))\ne.prop("PM2.5", 12, unitcode="GP").prop("PM10", 18, unitcode="GP")\ne.prop("NO2", 8, unitcode="GP").prop("O3", 83, unitcode="GP")\ne.rel("refDevice", "Device:AirProbe42")\nwith Client() as client:\n    client.upsert(e)\n```\n\nThe corresponding JSON-LD [payload](https://github.com/Orange-OpenSource/python-ngsild-client/blob/master/samples/gettingstarted.json) has been generated.\n\n## Documentation\n\nUser guide is available on [Read the Docs](https://ngsildclient.readthedocs.io/en/latest/index.html).\n\n## License\n\n[Apache 2.0](LICENSE)\n',
    'author': 'fbattello',
    'author_email': 'fabien.battello@orange.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Orange-OpenSource/python-ngsild-client',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

"""This module contains constants used in the model package.
"""

from enum import Enum
from typing import Union
from datetime import datetime, date, time, timezone
from zoneinfo import ZoneInfo
from geojson import Point, LineString, Polygon

from ..utils.sentinel import Sentinel

NgsiLocation = Union[tuple[int, int], Point]
"""A user type : either a tuple of two ints (lat, lon) or a GeoJson Point.
"""

NgsiGeometry = Union[NgsiLocation, LineString, Polygon]
"""A user type : Valid Geometries types for a NGSI-LD GeoProperty.
"""

NgsiDate = Union[str, datetime, date, time]
"""A user type : Valid Date types for a NGSI-LD TemporalProperty.
"""

CORE_CONTEXT = "https://uri.etsi.org/ngsi-ld/v1/ngsi-ld-core-context.jsonld"
"""The default NGSI-LD Core Context. Automatically set at entity creation time if no context is provided.
"""

META_ATTR_CONTEXT = "@context"
META_ATTR_UNITCODE = "unitCode"
META_ATTR_OBSERVED_AT = "observedAt"
META_ATTR_DATASET_ID = "datasetId"

NESTED = True

TZ_UTC = timezone.utc
TZ_WET = ZoneInfo("WET")  # UTC+1 i.e. Europe/Lisbon
TZ_CET = ZoneInfo("CET")  # UTC+2 i.e. Europe/Paris
TZ_FET = ZoneInfo("Europe/Minsk")  # UTC+3


class Auto(Sentinel):
    pass


class Rel(Enum):
    HAS_PART = "hasPart"
    HAS_DIRECT_PART = "hasDirectPart"
    IS_CONTAINED_IN = "isContainedIn"
    OBSERVED_BY = "observedBy"
    PROVIDED_BY = "providedBy"


class GeometryMetaAttr(Enum):
    LOCATION = "location"
    OBSERVATION_SPACE = "observationSpace"
    OPERATION_SPACE = "operationSpace"


class AttrType(Enum):
    PROP = "Property"
    TEMPORAL = "Property"  # Temporal Property
    GEO = "GeoProperty"
    REL = "Relationship"


class GeometryType(Enum):
    POINT = "Point"
    LINESTRING = "LineString"
    POLYGON = "Polygon"


class TemporalType(Enum):
    DATETIME = "DateTime"
    DATE = "Date"
    TIME = "Time"


SMARTDATAMODELS_BASEURL = "https://smart-data-models.github.io"  # Github Pages


class SmartDataModels:
    class SmartAgri:
        class Agrifood:
            AgriApp = f"{SMARTDATAMODELS_BASEURL}/dataModel.Agrifood/AgriApp/examples/example-normalized.jsonld"
            AgriCrop = f"{SMARTDATAMODELS_BASEURL}/dataModel.Agrifood/AgriCrop/examples/example-normalized.jsonld"
            AgriFarm = f"{SMARTDATAMODELS_BASEURL}/dataModel.Agrifood/AgriFarm/examples/example-normalized.jsonld"
            AgriGreenhouse = f"{SMARTDATAMODELS_BASEURL}/dataModel.Agrifood/AgriGreenhouse/examples/example-normalized.jsonld"
            AgriParcel = f"{SMARTDATAMODELS_BASEURL}/dataModel.Agrifood/AgriParcel/examples/example-normalized.jsonld"
            AgriParcelOperation = f"{SMARTDATAMODELS_BASEURL}/dataModel.Agrifood/AgriParcelOperation/examples/example-normalized.jsonld"
            AgriParcelRecord = f"{SMARTDATAMODELS_BASEURL}/dataModel.Agrifood/AgriParcelRecord/examples/example-normalized.jsonld"
            AgriPest = f"{SMARTDATAMODELS_BASEURL}/dataModel.Agrifood/AgriPest/examples/example-normalized.jsonld"
            AgriProductType = f"{SMARTDATAMODELS_BASEURL}/dataModel.Agrifood/AgriProductType/examples/example-normalized.jsonld"
            AgriSoil = f"{SMARTDATAMODELS_BASEURL}/dataModel.Agrifood/AgriSoil/examples/example-normalized.jsonld"
            Animal = f"{SMARTDATAMODELS_BASEURL}/dataModel.Agrifood/Animal/examples/example-normalized.jsonld"
            Compartment = f"{SMARTDATAMODELS_BASEURL}/dataModel.Agrifood/Compartment/examples/example-normalized.jsonld"
            Pen = f"{SMARTDATAMODELS_BASEURL}/dataModel.Agrifood/Pen/examples/example-normalized.jsonld"

        class Aquaculture:
            BreedingOperation = f"{SMARTDATAMODELS_BASEURL}/dataModel.Aquaculture/BreedingOperation/examples/example-normalized.jsonld"
            FishContainment = f"{SMARTDATAMODELS_BASEURL}/dataModel.Aquaculture/FishContainment/examples/example-normalized.jsonld"

        class Weather:
            SeaConditions = f"{SMARTDATAMODELS_BASEURL}/dataModel.Weather/SeaConditions/examples/example-normalized.jsonld"
            WeatherAlert = f"{SMARTDATAMODELS_BASEURL}/dataModel.Weather/WeatherAlert/examples/example-normalized.jsonld"
            WeatherForecast = f"{SMARTDATAMODELS_BASEURL}/dataModel.Weather/WeatherForecast/examples/example-normalized.jsonld"
            WeatherObserved = f"{SMARTDATAMODELS_BASEURL}/dataModel.Weather/WeatherObserved/examples/example-normalized.jsonld"

    class SmartCities:
        class Building:
            Building = f"{SMARTDATAMODELS_BASEURL}/dataModel.Building/Building/examples/example-normalized.jsonld"
            BuildingOperation = f"{SMARTDATAMODELS_BASEURL}/dataModel.Building/BuildingOperation/examples/example-normalized.jsonld"
            VibrationsObserved = f"{SMARTDATAMODELS_BASEURL}/dataModel.Building/VibrationsObserved/examples/example-normalized.jsonld"

        class Parking:
            OffStreetParking = f"{SMARTDATAMODELS_BASEURL}/dataModel.Parking/OffStreetParking/examples/example-normalized.jsonld"
            OnStreetParking = f"{SMARTDATAMODELS_BASEURL}/dataModel.Parking/OnStreetParking/examples/example-normalized.jsonld"
            ParkingAccess = f"{SMARTDATAMODELS_BASEURL}/dataModel.Parking/ParkingAccess/examples/example-normalized.jsonld"
            ParkingGroup = f"{SMARTDATAMODELS_BASEURL}/dataModel.Parking/ParkingGroup/examples/example-normalized.jsonld"
            ParkingSpot = f"{SMARTDATAMODELS_BASEURL}/dataModel.Parking/ParkingSpot/examples/example-normalized.jsonld"

        class ParksAndGarden:
            FlowerBed = f"{SMARTDATAMODELS_BASEURL}/dataModel.ParksAndGardens/FlowerBed/examples/example-normalized.jsonld"
            Garden = f"{SMARTDATAMODELS_BASEURL}/dataModel.ParksAndGardens/Garden/examples/example-normalized.jsonld"
            GreenSpaceRecord = f"{SMARTDATAMODELS_BASEURL}/dataModel.ParksAndGardens/GreenspaceRecord/examples/example-normalized.jsonld"

        class PointOfInterest:
            Beach = f"{SMARTDATAMODELS_BASEURL}/dataModel.PointOfInterest/Beach/examples/example-normalized.jsonld"
            Museum = f"{SMARTDATAMODELS_BASEURL}/dataModel.PointOfInterest/Museum/examples/example-normalized.jsonld"
            PointOfInterest = f"{SMARTDATAMODELS_BASEURL}/dataModel.PointOfInterest/PointOfInterest/examples/example-normalized.jsonld"
            Store = f"{SMARTDATAMODELS_BASEURL}/dataModel.PointOfInterest/Store/examples/example-normalized.jsonld"

        class Ports:
            BoatAuthorized = f"{SMARTDATAMODELS_BASEURL}/dataModel.Ports/BoatAuthorized/examples/example-normalized.jsonld"
            SeaportFacilities = f"{SMARTDATAMODELS_BASEURL}/dataModel.Ports/SeaportFacilities/examples/example-normalized.jsonld"
            BoatPlacesAvailable = f"{SMARTDATAMODELS_BASEURL}/dataModel.Ports/BoatPlacesAvailable/examples/example-normalized.jsonld"

        class StreetLightning:
            StreetLight = f"{SMARTDATAMODELS_BASEURL}/dataModel.Streetlighting/Streetlight/examples/example-normalized.jsonld"
            StreetLightControlCabinet = f"{SMARTDATAMODELS_BASEURL}/dataModel.Streetlighting/StreetlightControlCabinet/examples/example-normalized.jsonld"
            StreetLightGroup = f"{SMARTDATAMODELS_BASEURL}/dataModel.Streetlighting/StreetlightGroup/examples/example-normalized.jsonld"
            StreetLightModel = f"{SMARTDATAMODELS_BASEURL}/dataModel.Streetlighting/StreetlightModel/examples/example-normalized.jsonld"

        class Transportation:
            BikeHireDockingStation = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/BikeHireDockingStation/examples/example-normalized.jsonld"
            BikeLane = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/BikeLane/examples/example-normalized.jsonld"
            CityWork = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/CityWork/examples/example-normalized.jsonld"
            CrowFlowObserved = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/CrowdFlowObserved/examples/example-normalized.jsonld"
            EVChargingStation = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/EVChargingStation/examples/example-normalized.jsonld"
            FareCollectionSystem = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/FareCollectionSystem/examples/example-normalized.jsonld"
            ItemFlowObserved = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/ItemFlowObserved/examples/example-normalized.jsonld"
            RestrictedTrafficArea = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/RestrictedTrafficArea/examples/example-normalized.jsonld"
            RestrictionException = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/RestrictionException/examples/example-normalized.jsonld"
            Road = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/Road/examples/example-normalized.jsonld"
            RoadAccident = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/RoadAccident/examples/example-normalized.jsonld"
            RoadSegment = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/RoadSegment/examples/example-normalized.jsonld"
            SpecialRestriction = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/SpecialRestriction/examples/example-normalized.jsonld"
            TrafficFlowObserved = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/TrafficFlowObserved/examples/example-normalized.jsonld"
            TransportStation = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/TransportStation/examples/example-normalized.jsonld"
            Vehicle = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/Vehicle/examples/example-normalized.jsonld"
            VehicleModel = f"{SMARTDATAMODELS_BASEURL}/dataModel.Transportation/VehicleModel/examples/example-normalized.jsonld"

        class UrbanMobility:
            ArrivalEstimation = f"{SMARTDATAMODELS_BASEURL}/dataModel.UrbanMobility/ArrivalEstimation/examples/example-normalized.jsonld"
            GtfsAccessPoint = f"{SMARTDATAMODELS_BASEURL}/dataModel.UrbanMobility/GtfsAccessPoint/examples/example-normalized.jsonld"
            GtfsAgency = f"{SMARTDATAMODELS_BASEURL}/dataModel.UrbanMobility/GtfsAgency/examples/example-normalized.jsonld"
            GtfsCalendarDateRule = f"{SMARTDATAMODELS_BASEURL}/dataModel.UrbanMobility/GtfsCalendarDateRule/examples/example-normalized.jsonld"
            GtfsCalendarRule = f"{SMARTDATAMODELS_BASEURL}/dataModel.UrbanMobility/GtfsCalendarRule/examples/example-normalized.jsonld"
            GtfsFrequency = f"{SMARTDATAMODELS_BASEURL}/dataModel.UrbanMobility/GtfsFrequency/examples/example-normalized.jsonld"
            GtfsRoute = f"{SMARTDATAMODELS_BASEURL}/dataModel.UrbanMobility/GtfsRoute/examples/example-normalized.jsonld"
            GtfsService = f"{SMARTDATAMODELS_BASEURL}/dataModel.UrbanMobility/GtfsService/examples/example-normalized.jsonld"
            GtfsShape = f"{SMARTDATAMODELS_BASEURL}/dataModel.UrbanMobility/GtfsShape/examples/example-normalized.jsonld"
            GtfsStation = f"{SMARTDATAMODELS_BASEURL}/dataModel.UrbanMobility/GtfsStation/examples/example-normalized.jsonld"
            GtfsStop = f"{SMARTDATAMODELS_BASEURL}/dataModel.UrbanMobility/GtfsStop/examples/example-normalized.jsonld"
            GtfsStopTime = f"{SMARTDATAMODELS_BASEURL}/dataModel.UrbanMobility/GtfsStopTime/examples/example-normalized.jsonld"
            GtfsTransferRule = f"{SMARTDATAMODELS_BASEURL}/dataModel.UrbanMobility/GtfsTransferRule/examples/example-normalized.jsonld"
            GtfsTrip = f"{SMARTDATAMODELS_BASEURL}/dataModel.UrbanMobility/GtfsTrip/examples/example-normalized.jsonld"
            PublicTransportRoute = f"{SMARTDATAMODELS_BASEURL}/dataModel.UrbanMobility/PublicTransportRoute/examples/example-normalized.jsonld"
            PublicTransportStop = f"{SMARTDATAMODELS_BASEURL}/dataModel.UrbanMobility/PublicTransportStop/examples/example-normalized.jsonld"

        class WasteManagement:
            WasteContainer = f"{SMARTDATAMODELS_BASEURL}/dataModel.WasteManagement/WasteContainer/examples/example-normalized.jsonld"
            WasteContainerIsle = f"{SMARTDATAMODELS_BASEURL}/dataModel.WasteManagement/WasteContainerIsle/examples/example-normalized.jsonld"
            WasteContainerModel = f"{SMARTDATAMODELS_BASEURL}/dataModel.WasteManagement/WasteContainerModel/examples/example-normalized.jsonld"

        class Weather:
            SeaConditions = f"{SMARTDATAMODELS_BASEURL}/dataModel.Weather/SeaConditions/examples/example-normalized.jsonld"
            WeatherAlert = f"{SMARTDATAMODELS_BASEURL}/dataModel.Weather/WeatherAlert/examples/example-normalized.jsonld"
            WeatherForecast = f"{SMARTDATAMODELS_BASEURL}/dataModel.Weather/WeatherForecast/examples/example-normalized.jsonld"
            WeatherObserved = f"{SMARTDATAMODELS_BASEURL}/dataModel.Weather/WeatherObserved/examples/example-normalized.jsonld"



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

"""Sentinel util class.

See Also
--------
https://stackoverflow.com/questions/69239403/type-hinting-parameters-with-a-sentinel-value-as-the-default
https://www.python.org/dev/peps/pep-0661/
"""

from typing import Literal


class SentinelMeta(type):
    def __repr__(cls) -> str:
        return f"<{cls.__name__}>"

    def __bool__(cls) -> Literal[False]:
        return False


class Sentinel(metaclass=SentinelMeta):
    pass



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

from __future__ import annotations

import json
import requests
import logging

from copy import deepcopy
from functools import partialmethod
from dataclasses import dataclass

from datetime import datetime
from typing import overload, Any, Union, List, Optional, Callable
from rich import print_json

from .exceptions import *
from .constants import *
from .ngsidict import NgsiDict
from ngsildclient.utils import iso8601, url, is_interactive
from ngsildclient.utils.urn import Urn

logger = logging.getLogger(__name__)

"""This module contains the definition of the Entity class.
"""


class Entity:
    """The main goal of this class is to build, manipulate and represent a NGSI-LD compliant entity.

    The preferred constructor allows to create an entity from its NGSI-LD type and identifier (and optionally context),
    If no context is provided, it defaults to the NGSI-LD Core Context.
    The identifier can be written using the "long form" : the fully qualified urn string.
    It can also be shorten :
    - omit the urn prefix (scheme+nss)
    - omit the type inside the identifier, assuming the naming convention "urn:ngsi-ld:<type>:<remainder>"

    An alternate constructor allows to create an entity by just providing its id (and optionally context).
    In this case the type is inferred from the fully qualified urn string,
    assuming the naming convention "urn:ngsi-ld:<type>:<remainder>".

    The load() classmethod allows to load a NGSI-LD entity from a file or remotely through HTTP.
    For convenience some `SmartDataModels <https://smartdatamodels.org/>`_ examples are made available.

    Once initiated, we can build a complete NGSI-LD entity by adding attributes to the NGSI-LD entity.
    An attribute can be a Property, TemporalProperty, GeoProperty or RelationShip.
    Methods prop(), tprop(), gprop() and rel() are used to respectively build a Property, TemporalProperty,
    GeoProperty, and Relationship.
    Attributes can carry metadatas, such as "observedAt".
    Attributes can carry user data.

    Nested attributes are supported.
    Methods prop(), tprop(), gprop() are chainable, allowing to build nested properties.

    Dates and Datetimes are ISO8601.
    Helper functions are provided in the module utils.iso8601.
    Often a same date (the one when the measure/event happened) is used many times in the entity.
    When building an entity, the first time a datetime is used it is cached, then can be reused using "Auto".

    Given a NGSI-LD entity, many actions are possible :
    - access/add/remove/update attributes
    - access/update/remove values
    - print the content in the normalized or simplified (aka KeyValues) flavor
    - save the entity to a file
    - send it to the NGSI-LD Context Broker for creation/update (use the ngsildclient.api package)

    A NGSI-LD entity is backed by a NgsiDict object (a custom dictionary that inherits from the native Python dict).
    So if for any reasons you're stuck with the library and cannot achieve to build a NGSI-LD entity
    that fully matches your target datamodel, it's always possible to manipulate directly the underlying dictionary.

    Raises
    ------
    NgsiMissingIdError
        The identifier is missing
    NgsiMissingTypeError
        The type is missing
    NgsiMissingContextError
        The context is missing

    See Also
    --------
    model.NgsiDict : a custom dictionary that inherits from the native dict and provides primitives to build attributes
    api.client.Client : the NGSI-LD Context Broker client to interact with a Context Broker

    Example:
    --------
    >>> from datetime import datetime
    >>> from ngsildclient import *

    >>> # Create the entity
    >>> e = Entity("AirQualityObserved", "RZ:Obsv4567")

    >>> # Add a temporal property named dateObserved
    >>> # We could provide a string if preferred (rather than a datetime)
    >>> e.tprop("dateObserved", datetime(2018, 8, 7, 12))

    >>> # Add a property named NO2 with a pollutant concentration value and a metadata to indicate the unit (mg/m3)
    >>> # The accuracy property is nested
    >>> e.prop("NO2", 22, unitcode="GP", observedat=Auto).prop("accuracy", 0.95, NESTED)

    >>> # Add a relationship towards a POI NGSI-LD Entity
    >>> e.rel("refPointOfInterest", "PointOfInterest:RZ:MainSquare")

    >>> # Pretty-print to standard output
    >>> e.pprint()
    {
        "@context": [
            "https://uri.etsi.org/ngsi-ld/v1/ngsi-ld-core-context.jsonld"
        ],
        "id": "urn:ngsi-ld:AirQualityObserved:RZ:Obsv4567",
        "type": "AirQualityObserved",
        "dateObserved": {
            "type": "Property",
            "value": {
                "@type": "DateTime",
                "@value": "2018-08-07T12:00:00Z"
            }
        },
        "NO2": {
            "type": "Property",
            "value": 22,
            "unitCode": "GP",
            "observedAt": "2018-08-07T12:00:00Z",
            "accuracy": {
                "type": "Property",
                "value": 0.95
            }
        },
        "refPointOfInterest": {
            "type": "Relationship",
            "object": "urn:ngsi-ld:PointOfInterest:RZ:MainSquare"
        }
    }

    >>> # Update a property by overriding it
    >>> e.prop("dateObserved", iso8601.utcnow())

    >>> # Update a value using the dot notation
    >>> e["NO2.accuracy.value"] = 0.96

    >>> # Remove a property
    >>> e.rm("NO2.accuracy")
    """

    @dataclass
    class Settings:
        """The default settings used to build an Entity"""

        autoprefix: bool = True
        """A boolean to enable/disable the automatic insertion of the type into the identifier.
        Default is enabled.
        """

        strict: bool = False  # for future use
        autoescape: bool = True  # for future use
        f_print: Callable = print_json if is_interactive() else print

    globalsettings: Settings = Settings()

    @overload
    def __init__(self, type: str, id: str, *, ctx: list = [CORE_CONTEXT]):
        """Create a NGSI-LD compliant entity

        One can omit the urn and namespace, "urn:ngsi-ld:" will be added automatically.
        One can omit the type inside the identifier.

        By default, the constructor assumes the identifier naming convention "urn:ngsi-ld:<type>:<remainder>" and automatically
        insert the type into the identifier.
        The default behaviour can be disabled : Entity.globalsettings.autoprefix = False.


        Parameters
        ----------
        type : str
            entity type
        id : str
            entity identifier
        ctx : list, optional
            the NGSI-LD context, by default the NGSI-LD Core Context

        Example:
        --------
        >>> from ngsildclient.model.entity import Entity
        >>> e1 = Entity("AirQualityObserved", "urn:ngsi-ld:AirQualityObserved:RZ:Obsv4567") # long form
        >>> e2 = Entity("AirQualityObserved", "AirQualityObserved:RZ:Obsv4567") # omit scheme + nss
        >>> e3 = Entity("AirQualityObserved", "RZ:Obsv4567") # omit scheme + nss + type
        >>> print(e1 == e2 == e3)
        True
        >>> e1.pprint()
        {
            "@context": [
                "https://uri.etsi.org/ngsi-ld/v1/ngsi-ld-core-context.jsonld"
            ],
            "id": "urn:ngsi-ld:AirQualityObserved:RZ:Obsv4567",
            "type": "AirQualityObserved"
        }
        """
        ...

    @overload
    def __init__(self, id: str, *, ctx: list = [CORE_CONTEXT]):
        """Create a NGSI-LD compliant entity.

        Type is inferred from the fully qualified identifier.
        Works only if your identifiers follow the naming convention "urn:ngsi-ld:<type>:<remainder>"

        Parameters
        ----------
        id : str
            entity identifier (fully qualified urn)
        context : list, optional
            the NGSI-LD context, by default the NGSI-LD Core Context

        Example:
        --------
        >>> from ngsildclient.model.entity import Entity
        >>> e = Entity("urn:ngsi-ld:AirQualityObserved:RZ:Obsv4567")
        >>> e.pprint()
        {
            "@context": [
                "https://uri.etsi.org/ngsi-ld/v1/ngsi-ld-core-context.jsonld"
            ],
            "id": "urn:ngsi-ld:AirQualityObserved:RZ:Obsv4567",
            "type": "AirQualityObserved"
        }
        """
        ...

    def __init__(
        self,
        arg1: str = None,
        arg2: str = None,
        *,
        ctx: list = [CORE_CONTEXT],
        payload: dict = None,
        autoprefix: Optional[bool] = None,
    ):
        logger.debug(f"{arg1=} {arg2=}")

        if autoprefix is None:
            autoprefix = Entity.globalsettings.autoprefix

        self._lastprop: NgsiDict = None
        self._anchored: bool = False

        if payload is not None:  # create Entity from a dictionary
            if not payload.get("id", None):
                raise NgsiMissingIdError()
            if not payload.get("type", None):
                raise NgsiMissingTypeError()
            if not payload.get("@context", None):
                raise NgsiMissingContextError()
            self._payload: NgsiDict = NgsiDict(payload)
            return

        # create a new Entity using its id and type

        if arg2:
            type, id = arg1, arg2
        else:
            type, id = None, arg1

        dt = iso8601.extract(id)

        if type is None:  # try to infer type from the fully qualified identifier
            id = Urn.prefix(id)
            urn = Urn(id)
            if (type := urn.infertype()) is None:
                raise NgsiMissingTypeError(f"{urn.fqn=}")
        else:  # type is not None
            autoprefix &= not Urn.is_prefixed(id)
            if autoprefix:
                bareid = Urn.unprefix(id)
                prefix = f"{type}:"
                if not bareid.startswith(prefix):
                    id = prefix + bareid
            id = Urn.prefix(id)  # set the prefix "urn:ngsi-ld:" if not already done
            urn = Urn(id)

        self._payload: NgsiDict = NgsiDict(
            {"@context": ctx, "id": urn.fqn, "type": type},
            dtcached=dt if dt else iso8601.utcnow(),
        )

    @classmethod
    def from_dict(cls, payload: dict):
        """Create a NGSI-LD entity from a dictionary.

        The input dictionary must at least contain the 'id', 'type' and '@context'.
        This method assumes that the input dictionary matches a valid NGSI-LD structure.

        Parameters
        ----------
        payload : dict
            The given dictionary.

        Returns
        -------
        Entity
            The result Entity instance
        """
        return cls(payload=payload)

    @classmethod
    def duplicate(cls, entity: Entity) -> Entity:
        """Duplicate a given entity.

        Parameters
        ----------
        entity : Entity
            The input Entity

        Returns
        -------
        Entity
            The output entity
        """
        new = deepcopy(entity)
        return new

    def copy(self) -> Entity:
        """Duplicates the entity

        Returns
        -------
        Entity
            The new entity
        """
        return Entity.duplicate(self)

    @property
    def id(self):
        return self._payload["id"]

    @id.setter
    def id(self, eid: str):
        self._payload["id"] = eid

    @property
    def type(self):
        return self._payload["type"]

    @type.setter
    def type(self, etype: str):
        self._payload["type"] = etype

    @property
    def context(self):
        return self._payload["@context"]

    @context.setter
    def context(self, ctx: list):
        self._payload["@context"] = ctx

    def __getitem__(self, item):
        return self._payload.__getitem__(item)

    def __setitem__(self, key, item):
        self._payload.__setitem__(key, item)

    def __delitem__(self, key):
        self._payload.__delitem__(key)
        return self

    def last(self, name: str):
        item = self[name]
        if not isinstance(item, List):
            raise ValueError("Item should be a list")
        return deepcopy(item[-1])

    def append(self, name: str, value: NgsiDict):
        item = self[name]
        if not isinstance(item, List):
            raise ValueError("Item should be a list")
        item.append(value)
        return deepcopy(item[-1])

    def anchor(self):
        """Set an anchor.

        Allow to specify that the last property is used as an anchor.
        Once the anchor property is specified, new properties are attached to the anchor property.

        Parameters
        ----------
        entity : Entity
            The input Entity

        Returns
        -------
        Entity
            The output entity

        Example:
        --------
        Here an anchor is set at the "availableSpotNumber" property.
        Hence the "reliability" and "providedBy" properties are attached to (nested in) the "availableSpotNumber" property.
        Without anchoring, the "reliability" and "providedBy" properties would apply to the entity's root.

        >>> from ngsildclient.model.entity import Entity
        >>> e = Entity("OffStreetParking", "Downtown1")
        >>> e.prop("availableSpotNumber", 121, observedat=datetime(2017, 7, 29, 12, 5, 2)).anchor()
        >>> e.prop("reliability", 0.7).rel("providedBy", "Camera:C1").unanchor()
        >>> e.pprint()
        {
            "@context": [
                "http://uri.etsi.org/ngsi-ld/v1/ngsi-ld-core-context.jsonld",
                "http://example.org/ngsi-ld/parking.jsonld"
            ],
            "id": "urn:ngsi-ld:OffStreetParking:Downtown1",
            "type": "OffStreetParking",
            "availableSpotNumber": {
                "type": "Property",
                "value": 121,
                "observedAt": "2017-07-29T12:05:02Z",
                "reliability": {
                    "type": "Property",
                    "value": 0.7
                },
                "providedBy": {
                    "type": "Relationship",
                    "object": "urn:ngsi-ld:Camera:C1"
                }
            }
        }
        """
        self._anchored = True
        return self

    def unanchor(self):
        """Remove the anchor.

        See Also
        --------
        Entity.anchor()
        """

        self._anchored = False
        return self

    def _update_entity(self, attrname: str, property: NgsiDict, nested: bool = False):
        nested |= self._anchored
        if nested and self._lastprop is not None:
            # update _lastprop only if not anchored
            self._lastprop[attrname] = property
            if not self._anchored:
                self._lastprop = property
        else:
            self._lastprop = self._payload[attrname] = property

    def prop(
        self,
        name: str,
        value: Any,
        nested: bool = False,
        *,  # keyword-only arguments after this
        unitcode: str = None,
        observedat: Union[str, datetime] = None,
        datasetid: str = None,
        userdata: NgsiDict = NgsiDict(),
        escape: bool = False,
    ) -> Entity:
        """Build a Property.

        Build a property and attach it to the current entity.
        One can chain prop(),tprop(), gprop(), rel() methods to build nested properties.

        Parameters
        ----------
        name : str
            the property name
        value : Any
            the property value
        observedat : Union[str, datetime], optional
            observetAt metadata, timestamp, ISO8601, UTC, by default Noneself._update_entity(name, property, nested)
        userdata : NgsiDict, optional
            a dict or NgsiDict containing user data, i.e. userdata={"reliability": 0.95}, by default NgsiDict()
        escape : bool, optional
            if set escape the string value (useful if contains forbidden characters), by default False

        Returns
        -------
        Entity
            The updated entity

        Example:
        --------
        >>> from ngsildclient.model.entity import Entity
        >>> e = Entity("AirQualityObserved", "RZ:Obsv4567")
        >>> e.prop("NO2", 22, unitcode="GP") # basic property
        {'type': 'Property', 'value': 22, 'unitCode': 'GP'}
        >>> e.prop("PM10", 18, unitcode="GP").prop("reliability", 0.95) # chain methods to obtain a nested property
        {'type': 'Property', 'value': 0.95}
        >>> e.pprint()
        {
        "@context": [
            "https://uri.etsi.org/ngsi-ld/v1/ngsi-ld-core-context.jsonld"
        ],
        "id": "urn:ngsi-ld:AirQualityObserved:RZ:Obsv4567",
        "type": "AirQualityObserved",
        "NO2": {
            "type": "Property",
            "value": 22,
            "unitCode": "GP"
        },
        "PM10": {
            "type": "Property",
            "value": 18,
            "unitCode": "GP",
            "reliability": {
                "type": "Property",
                "value": 0.95
            }
        }
        """
        property = self._payload._build_property(value, unitcode, observedat, datasetid, userdata, escape)
        self._update_entity(name, property, nested)
        return self

    def addr(self, value: str):
        return self.prop("address", value)

    def gprop(
        self,
        name: str,
        value: NgsiGeometry,
        nested: bool = False,
        observedat: Union[str, datetime] = None,
        datasetid: str = None,
    ) -> Entity:
        """Build a GeoProperty.

        Build a GeoProperty and attach it to the current entity.
        One can chain prop(),tprop(), gprop(), rel() methods to build nested properties.

        Parameters
        ----------
        name : str
            the property name
        value : NgsiGeometry
            the property value

        Returns
        -------
        Entitydatasetid
            The updated entity

        Example:
        --------
        >>> from ngsildclient.model.entity import Entity
        >>> e = Entity("PointOfInterest", "RZ:MainSquare")
        >>> e.prop("description", "Beach of RZ")
        >>> e.gprop("location", (44, -8))
        >>> e.pprint()
        {
            "@context": [
                "https://uri.etsi.org/ngsi-ld/v1/ngsi-ld-core-context.jsonld"
            ],
            "id": "urn:ngsi-ld:PointOfInterest:RZ:MainSquare",
            "type": "PointOfInterest",
            "description": {
                "type": "Property",
                "value": "Beach of RZ"
            },
            "location": {datasetid
                "type": "GeoProperty",
                "value": {
                "type": "Point",
                "coordinates": [
                    -8,datasetid
            }
        }
        """
        property = self._payload._build_geoproperty(value, observedat, datasetid)
        self._update_entity(name, property, nested)
        return self

    loc = partialmethod(gprop, "location")
    """ A helper method to set the frequently used "location" geoproperty.

    entity.loc((44, -8)) is a shorcut for entity.gprop("location", (44, -8))
    """

    def tprop(self, name: str, value: NgsiDate = None, nested: bool = False) -> Entity:
        """Build a TemporalProperty.

        Build a TemporalProperty and attach it to the current entity.
        One can chain prop(),tprop(), gprop(), rel() methoddatasetids to build nested properties.

        Parameters
        ----------
        name : str
            the property name
        value : NgsiDate
            the property value, utcnow() if None

        Returns
        -------
        Entity
            The updated entity

        Example:
        --------
        >>> from datetime import datetime
        >>> from ngsildclient.model.entity import Entity
        >>> e = Entity("AirQualityObserved", "RZ:Obsv4567")
        >>> e.tprop("dateObserved", datetime(2018, 8, 7, 12))
        >>> e.pprint()
        {
            "@context": [datasetid
                "https://uri.etsi.org/ngsi-ld/v1/ngsi-ld-core-context.jsonld"
            ],
            "id": "urn:ngsi-ld:AirQualityObserved:RZ:Obsv4567",
            "type": "AirQualityObserved",
            "dateObserved": {
                "type": "Property",
                "value": {
                    "@type": "DateTime",
                    "@value": "2018-08-07T12:00:00Z"
                }
            }
        }
        """
        if value is None:
            value = self._payload._dtcached
        property = self._payload._build_temporal_property(value)
        self._update_entity(name, property, nested)
        return self

    obs = partialmethod(tprop, "dateObserved")
    """ A helper method to set the frequently used "dateObserved" property.

    entity.obs("2022-01-12T12:54:38Z") is a shorcut for entity.tprop("dateObserved", "2022-01-12T12:54:38Z")
    """

    def rel(
        self,
        name: Union[Rel, str],
        value: Union[str, List[str], Entity, List[Entity]],
        nested: bool = False,
        *,
        observedat: Union[str, datetime] = None,
        datasetid: str = None,
        userdata: NgsiDict = NgsiDict(),
    ) -> Entity:
        """Build a Relationship Property.

        Build a Relationship Property and attach it to the current entity.
        One can chain prop(),tprop(), gprop(), rel() methods to build nested properties.

        Parameters
        ----------
        name : str
            the property name
        value : str
            the property value

        Returns
        -------
        Entity
            The updated entity

        Example:
        --------
        >>> from ngsildclient.model.entity import Entity
        >>> e = Entity("AirQualityObserved", "RZ:Obsv4567")
        >>> e.rel("refPointOfInterest", "PointOfInterest:RZ:MainSquare")
        >>> e.pprint()
        {
            "@context": [
                "https://uri.etsi.org/ngsi-ld/v1/ngsi-ld-core-context.jsonld"
            ],
            "id": "urn:ngsi-ld:AirQualityObserved:RZ:Obsv4567",
            "type": "AirQualityObserved",
            "refPointOfInterest": {
                "type": "Relationship",
                "object": "urn:ngsi-ld:PointOfInterest:RZ:MainSquare"
            }
        }
        """
        if isinstance(name, Rel):
            name = name.value

        if isinstance(value, List):
            property = self._payload._m_build_relationship(value, observedat, datasetid, userdata)
        else:
            property = self._payload._build_relationship(value, observedat, datasetid, userdata)
        self._update_entity(name, property, nested)

        return self

    def __eq__(self, other: Entity):
        if other.__class__ is not self.__class__:
            return NotImplemented
        return self._payload == other._payload

    def __repr__(self):
        return self._payload.__repr__()

    def arrayify(self, name: str) -> NgsiDict:
        prop = self[name]
        if isinstance(prop, List):
            return None
        self[name] = [prop]
        return deepcopy(prop)

    def unarrayify(self, name: str) -> NgsiDict:
        prop = self[name]
        if not isinstance(prop, List):
            return None
        lastprop = prop[-1]
        self[name] = lastprop
        return lastprop

    def to_dict(self, kv=False) -> NgsiDict:
        """Returns the entity as a dictionary.

        The returned type is NgsiDict, fully compatible with a native dict.

        Parameters
        ----------
        kv : bool, optional
            KeyValues format (aka simplified representation), by default False

        Returns
        -------
        NgsiDict
            The underlying native Python dictionary
        """
        return self._to_keyvalues() if kv else self._payload

    def _to_keyvalues(self) -> NgsiDict:
        """Compute a NgsiDict that contains only the highest-level of information.

        Returns
        -------
        NgsiDict
            The simplified representation
        """
        d = NgsiDict()
        for k, v in self._payload.items():
            if isinstance(v, dict):
                if v["type"] == AttrType.PROP.value:  # apply to Property and TemporalProperty
                    value = v["value"]
                    if isinstance(value, dict):
                        value = value.get("@value", value)  # for Temporal Property only
                    d[k] = value
                elif v["type"] == AttrType.GEO.value:
                    d[k] = v["value"]
                elif v["type"] == AttrType.REL.value:
                    d[k] = v["object"]
            else:
                d[k] = v
        return d

    def to_json(self, kv=False, *args, **kwargs) -> str:
        """Returns the entity as JSON.

        Parameters
        ----------
        kv : bool, optional
            KeyValues format (aka simplified representation), by default False

        Returns
        -------
        str
            The JSON content
        """
        payload: NgsiDict = self.to_dict(kv)
        return payload.to_json(*args, **kwargs)

    def pprint(self, kv=False, *args, **kwargs):
        """Pretty-print the entity to the standard ouput.

        Parameters
        ----------
        kv : bool, optional
            KeyValues format (aka simplified representation), by default False
        """
        Entity.globalsettings.f_print(self.to_json(kv, indent=2, *args, **kwargs))

    @classmethod
    def load(cls, filename: str):
        """Load an Entity from a JSON file, locally from the filesystem or remotely through HTTP.

        For convenience some `SmartDataModels <https://smartdatamodels.org/>`_ examples are made available thanks to the Smart Data Models initiative.
        You can benefit from autocompletion to navigate inside the available datamodels.

        Parameters
        ----------
        filename : str
            If filename corresponds to an URL, the JSON file is downloaded from HTTP.
            Else it is retrieved locally from the filesystem.

        Returns
        -------
        Entity
            The Entity instance

        See Also
        --------
        model.constants.SmartDataModels

        Example:
        --------
        >>> from ngsildclient import *
        >>> e = Entity.load(SmartDatamodels.SmartCities.Weather.WeatherObserved)
        """
        if url.isurl(filename):
            resp = requests.get(filename)
            payload = resp.json()
        else:
            with open(filename, "r") as fp:
                payload = json.load(fp)
        if isinstance(payload, List):
            return [cls.from_dict(x) for x in payload]
        return cls.from_dict(payload)

    @classmethod
    def load_batch(cls, filename: str):
        """Load a batch of entities from a JSON file.

        Parameters
        ----------
        filename : str
            The input file must contain a JSON array

        Returns
        -------
        List[Entity]
            A list of entities

        Example:
        --------
        >>> from ngsildclient import *
        >>> rooms = Entity.load_batch("/tmp/rooms_all.jsonld")
        """
        with open(filename, "r") as fp:
            payload = json.load(fp)
        if not isinstance(payload, List):
            raise ValueError("The JSON payload MUST be an array")
        return [cls.from_dict(x) for x in payload]

    def save(self, filename: str, *, indent: int = 2):
        """Save the entity to a file.

        Parameters
        ----------
        filename : str
            Name of the output file
        indent : int, optional
            identation size (number of spaces), by default 2
        """
        with open(filename, "w") as fp:
            json.dump(self._payload, fp, default=str, ensure_ascii=False, indent=indent)

    @classmethod
    def save_batch(cls, entities: List[Entity], filename: str, *, indent: int = 2):
        """Save a batch of entities to a JSON file.

        Parameters
        ----------
        entities: List[Entity]
            Batch of entities to be saved

        filename : str
            If filename corresponds to an URL, the JSON file is downloaded from HTTP.
            Else it is retrieved locally from the filesystem.

        Example:
        --------
        >>> from ngsildclient import *
        >>> rooms = [Entity("Room", "Room1"), Entity("Room", "Room2")]
        >>> Entity.save_batch(rooms, "/tmp/rooms_all.jsonld")
        """
        payload = [x._payload for x in entities]
        with open(filename, "w") as fp:
            json.dump(payload, fp, default=str, ensure_ascii=False, indent=indent)



#!/usr/bin/env python3

# Software Name: ngsildclient
# SPDX-FileCopyrightText: Copyright (c) 2021 Orange
# SPDX-License-Identifier: Apache 2.0
#
# This software is distributed under the Apache 2.0;
# see the NOTICE file for more details.
#
# Author: Fabien BATTELLO <fabien.battello@orange.com> et al.

"""
This module contains helper functions to URI/URN in the NGSI-LD namespace.

References
----------
.. [2] ETSI, 2021. "NGSI-LD namespace" in Context Information Management (CIM); NGSI-LD API
    ETSI GS CIM 009 V1.4.2, Annex A.3, p.183, 2021-04.
"""

import re
import logging

from typing import overload, Optional

NID_PATTERN = re.compile(r"^[0-9a-zA-Z\-]+$")
"""Regex pattern that matches a valid namespace identifier (`regex.Pattern`).
"""

URN_PATTERN = re.compile(r"^urn:(?P<nid>[0-9a-zA-Z\-]+):(?P<nss>.+)$")
"""Regex pattern that extracts NID and NSS from a full valid urn string(`regex.Pattern`).
"""

ENTITY_TYPE_PATTERN = re.compile(r"^([^:]+):.*")
"""Regex pattern that extracts type from a NGSI-LD NSS following naming convention.(`regex.Pattern`).
"""

logger = logging.getLogger(__name__)


class UrnError(Exception):
    """Exception raised when parsing invalid urn/uri."""

    pass


class Urn:
    """Helper class to handle NGSI-LD urn/uri and allow to work with unprefixed strings."""

    DEFAULT_NID = "ngsi-ld"
    # """Default NGSI-LD namespace value
    # """

    @overload
    def __init__(self, fqn: str) -> None:
        """Construct by providing the fully qualified string.

        Parameters
        ----------
        fqn : str
            the fully qualified name, starting with `urn:`
        """
        ...

    @overload
    def __init__(self, *, nss: str, nid: str = DEFAULT_NID) -> None:
        """Construct by specifying the URN parts.

        Parameters
        ----------
        nss : str
            the namespace specific string
        nid : str, optional
            the namespace identifier, by default DEFAULT_NID
        """
        ...

    def __init__(
        self, fqn: str = None, *, nss: str = None, nid: str = DEFAULT_NID
    ) -> None:
        self.scheme = "urn"
        if nss and nid:
            self.nss = nss
            self.nid = nid
        elif fqn:
            m = URN_PATTERN.match(fqn)
            if m is None:
                raise UrnError(f"Bad urn format : {fqn}")
            d = m.groupdict()
            self.nss = d["nss"]
            self.nid = d["nid"]
        else:
            raise ValueError("Wrong arguments")

    @property
    def fqn(self) -> str:
        """Returns the fully-qualified name

        Returns
        -------
        str
            the fully qualified name

        Examples
        --------
        >>> from ngsildclient.utils.urn import Urn
        >>> urn = Urn(nss="AirQualityObserved:RZ:Obsv4567")
        >>> print(urn.fqn)
        urn:ngsi-ld:AirQualityObserved:RZ:Obsv4567
        """
        return f"{self.scheme}:{self.nid}:{self.nss}"

    def __eq__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        return self.nid == other.nid and self.nss == other.nss

    def __repr__(self) -> str:
        return self.fqn

    def infertype(self) -> Optional[str]:
        """Infer type.

        Work only when following the naming convention `urn:ngsi-ld:<type>:...`

        Returns
        -------
        Optional[str]
            the inferred NGSI LD type if found
        """
        m = ENTITY_TYPE_PATTERN.match(self.nss)
        return m.group(1) if m else None

    @staticmethod
    def is_valid_nid(nid: str) -> bool:
        """Check whether the given nid is a valid one

        Parameters
        ----------
        nid : str
            the nid string to be checked

        Returns
        -------
        bool
            True if if the nid is valid

        Examples
        --------
        >>> from ngsildclient.utils.urn import Urn
        >>> print(Urn.is_valid_nid("ngsi-ld"))
        True
        >>> print(Urn.is_valid_nid("ngsi+ld"))
        False
        """
        m = NID_PATTERN.match(nid)
        return m is not None

    @staticmethod
    def is_prefixed(value: str) -> bool:
        """Check whether the given string is prefixed (URN scheme+NID)

        Parameters
        ----------
        value : str
            the string value to be checked

        Returns
        -------
        bool
            True if the string value starts with the NGSI-LD u`urn:ngsi-ld:`

        Examples
        --------
        >>> from ngsildclient.utils.urn import Urn
        >>> print(Urn.is_prefixed("urn:ngsi-ld:AirQualityObserved:RZ:Obsv4567"))
        True
        """
        return value.startswith("urn:ngsi-ld:")

    @staticmethod
    def prefix(value: str) -> str:
        """Prefix a string with URN scheme+NID

        Parameters
        ----------
        value : str
            the string to be prefixed

        Returns
        -------
        str
            the string prefixed (if not already prefixed)

        Examples
        --------
        >>> from ngsildclient.utils.urn import Urn
        >>> print(Urn.prefix("AirQualityObserved:RZ:Obsv4567"))
        urn:ngsi-ld:AirQualityObserved:RZ:Obsv4567
        """
        if value is None:
            return None
        return value if Urn.is_prefixed(value) else f"urn:ngsi-ld:{value}"

    @staticmethod
    def unprefix(value: str) -> str:
        """Remove the prefix (URN scheme+NID)

        Parameters
        ----------
        value : str
            the string to be unprefixed

        Returns
        -------
        str
            the string without the prefix

        Examples
        --------
        >>> from ngsildclient.utils.urn import Urn
        >>> print(Urn.unprefix("urn:ngsi-ld:AirQualityObserved:RZ:Obsv4567"))
        AirQualityObserved:RZ:Obsv4567
        """
        if value is None:
            return None
        return Urn(value).nss if Urn.is_prefixed(value) else value
